"""Basic tests of simple core functionality."""

from pathlib import Path

import common
import pytest

# fmt: off
#: Protein-nucleic acid complexes
PROTEIN_NUCLEIC_SET = {"4UN3"}
#: Proteins without nucleic acids
PROTEIN_ONLY_SET = {"1K1I", "1FAS", "5D8V"}
#: DNA without protein
DNA_ONLY_SET = {"1NAJ", "7BNA"}
#: RNA without protein
RNA_ONLY_SET = {"1AQO", "4E8M"}
#: Nucleic acids without proteins
NUCLEIC_ONLY_SET = DNA_ONLY_SET | RNA_ONLY_SET
#: Basic test set
SHORT_SET = PROTEIN_NUCLEIC_SET | PROTEIN_ONLY_SET | NUCLEIC_ONLY_SET
#: Long test set of known good structures
LONG_PROTEIN_SET = {
    "5DV8", "1AFS", "1E7G", "3H0X", "1DUP", "4RWU", "3JTE", "3CP1", "1TG0",
    "1USM", "2OO2", "2J6B", "2SFA", "3US6", "3CSR", "3V1A", "3LAA", "3LS0",
    "2OP6", "2NWD", "3ONJ", "1Z0P", "4YPC", "3CQT", "1YQB", "1OX3", "2P5K",
    "1R7J", "1H75", "1YIB", "4Q2Q", "4GQM", "1UG4", "4TTW", "2JIC", "4GS3",
    "3I1G", "4PGR", "1HOE", "2CKX", "1HUF", "1BM8", "2G7O", "1OA4", "3RJS",
    "1TEN", "2HC8", "3KZD", "3FH2", "1X91", "3T8R", "3B7H", "4N6T", "1J2A",
    "3QC7", "4S11", "3NGP", "2IGD", "2NYC", "1NXO", "4OSN", "3SO2", "2J70",
    "3ZSL", "2IVY", "2NSN", "4IL7", "1NH9", "2EBB", "2REY", "2JCP", "2CMP",
    "3ZHI", "2YXF", "1TUD", "4ZQA", "1X3O", "4O7Q", "3I35", "3CE7", "2CJJ",
    "2QR3", "1TGN", "3KT2", "1YZM", "1FAS", "2X4J", "1TIG", "1L2P", "4EO0",
    "2IBL", "1JQ0", "2QVO", "4G3O", "2ZQE", "2FD4", "1DF4", "2PMR", "2PII",
    "3HRO", "2D8E", "2PCY", "4GMQ", "1J27", "4CVD", "1OPC", "1P1L", "1I2T",
    "4NPN", "2J9V", "2CWR", "1ZX6", "4L9E", "1RJ1", "3ISU", "3RFI", "4LJ1",
    "3TSV", "3HRN", "2HP7", "2SGA", "3C97", "1R5Q", "1NKD", "2IWN", "1TQG",
    "3QE1", "1X6J", "1PZC", "2VWR", "1PDO", "2BYG", "1ZZK", "2YGS", "2ERW",
    "4HTI", "3LLB", "2FI9", "1P9I", "4I2T", "2GUS", "3T1S", "1V2Z", "2QBV",
    "3ZZP", "1PHT", "4HS2", "1USE", "5EE2", "4YKD", "4CFI", "1WOU", "3RFF",
    "1NOA", "2CGQ", "2TGI", "2R6Q", "1IS5", "4CRH", "1B0X", "2FWG", "2HDZ",
    "3WFV", "4AXT", "2ERA", "4ZMK", "1ZVA", "3S0A", "2PTV", "2FRG", "4POY",
    "1PGX", "1LN4", "1SJV", "2ESK", "2PND", "1XAK", "3NX6", "1TVQ", "3RVC",
    "1RIS", "1AIL", "3IDW", "1KFN", "3LMO", "1OPS", "2H2C", "3DVI", "2GZV",
    "2XXC", "1ZLM", "2OCH",
}
LONG_NUCLEIC_SET = {"5V0O"}
LONG_SET = LONG_PROTEIN_SET | LONG_NUCLEIC_SET
#: Tests that should fail (broken backbones)
BROKEN_SET = {"1EJG", "3U7T", "4MGP", "2V75"}
#: Tests for naming residues with different protonation states
NAMING_TESTS = []
for ff in ["CHARMM", "AMBER", "PARSE", "SWANSON", "TYL06"]:
    for pdb in ["1AJJ", "1BX8"]:
        for pH in [2, 7, 14]:
            options = {"pdb": pdb, "ff": ff, "pH": pH}
            if ff == "CHARMM":
                NAMING_TESTS.append(pytest.param(options, marks=pytest.mark.xfail(reason="CHARMM force field is broken!")))
            else:
                NAMING_TESTS.append(pytest.param(options))
# fmt: on


@pytest.mark.parametrize("input_pdb", list(SHORT_SET), ids=str)
def test_short_pdb(input_pdb, tmp_path):
    """Non-regression tests on short list of PDB-format biomolecules."""
    args = "--log-level=INFO --ff=AMBER --drop-water --apbs-input=apbs.in"
    output_pqr = Path(input_pdb).stem + ".pqr"
    common.run_pdb2pqr_for_tests(
        args=args,
        input_pdb=input_pdb,
        output_pqr=output_pqr,
        tmp_path=tmp_path,
    )


@pytest.mark.parametrize("input_pdb", list(SHORT_SET), ids=str)
def test_basic_cif(input_pdb, tmp_path):
    """Non-regression tests on short list of CIF-format biomolecules."""
    args = "--log-level=INFO --ff=AMBER --drop-water --apbs-input=apbs.in"
    output_pqr = Path(input_pdb).stem + ".cif"
    common.run_pdb2pqr_for_tests(
        args=args,
        input_pdb=input_pdb,
        output_pqr=output_pqr,
        tmp_path=tmp_path,
    )


@pytest.mark.parametrize("input_pdb", list(LONG_SET), ids=str)
@pytest.mark.long_test
def test_long_pdb(input_pdb, tmp_path):
    """Non-regression tests on short list of PDB-format biomolecules."""
    args = "--log-level=INFO --ff=AMBER --drop-water --apbs-input=apbs.in"
    output_pqr = Path(input_pdb).stem + ".pqr"
    common.run_pdb2pqr_for_tests(
        args=args,
        input_pdb=input_pdb,
        output_pqr=output_pqr,
        tmp_path=tmp_path,
    )


@pytest.mark.parametrize("input_pdb", list(BROKEN_SET), ids=str)
@pytest.mark.xfail
def test_broken_backbone(input_pdb, tmp_path):
    """Test graceful failure of optimization with missing backbone atoms."""
    args = "--log-level=INFO --ff=AMBER --drop-water"
    output_pqr = Path(input_pdb).stem + ".pqr"
    common.run_pdb2pqr_for_tests(
        args=args,
        input_pdb=input_pdb,
        output_pqr=output_pqr,
        tmp_path=tmp_path,
    )


@pytest.mark.parametrize(
    "input_pdb, expected_pqr",
    [pytest.param("cterm_hid.pdb", "cterm_hid_out.pqr", id="C-terminal HID")],
)
def test_protonated_terminals(input_pdb, expected_pqr, tmp_path):
    """Tests for terminal residue protonation."""
    args = "--log-level=INFO --ff=AMBER --ffout AMBER"
    output_pqr = Path(input_pdb).stem + ".pqr"
    common.run_pdb2pqr_for_tests(
        args=args,
        input_pdb=common.DATA_DIR / input_pdb,
        output_pqr=output_pqr,
        expected_pqr=common.DATA_DIR / expected_pqr,
        tmp_path=tmp_path,
        compare_resnames=True,
    )


@pytest.mark.parametrize(
    "input_pdb, expected_pqr",
    [
        pytest.param(
            "5vav_cyclic_peptide.pdb",
            "5vav_cyclic_peptide_out.pqr",
            id="Cyclic peptide",
        )
    ],
)
def test_cyclic_peptide(input_pdb, expected_pqr, tmp_path):
    """Tests for cyclic peptide protonation."""
    args = "--log-level=INFO --ff=AMBER --ffout AMBER"
    output_pqr = Path(input_pdb).stem + ".pqr"
    common.run_pdb2pqr_for_tests(
        args=args,
        input_pdb=common.DATA_DIR / input_pdb,
        output_pqr=output_pqr,
        expected_pqr=common.DATA_DIR / expected_pqr,
        tmp_path=tmp_path,
        compare_resnames=True,
    )


@pytest.mark.parametrize("naming_test", NAMING_TESTS, ids=str)
def test_ph_naming(naming_test, tmp_path):
    """Non-regression tests on naming schemes at different pH values."""
    args = "--log-level=INFO --ff=AMBER --drop-water --apbs-input=apbs.in"
    input_pdb = naming_test["pdb"] + ".pdb"
    output_pqr = Path(input_pdb).stem + ".pqr"
    expected_pqr = (
        f"{naming_test['pdb']}_pH{naming_test['pH']}_{naming_test['ff']}.pqr"
    )
    args = (
        f"--log-level=INFO --ff={naming_test['ff']} --ffout={naming_test['ff']} "
        f"--drop-water --whitespace --with-ph={naming_test['pH']} "
        f"--titration-state-method=propka"
    )
    common.run_pdb2pqr_for_tests(
        args=args,
        input_pdb=common.DATA_DIR / input_pdb,
        output_pqr=output_pqr,
        expected_pqr=common.DATA_DIR / expected_pqr,
        tmp_path=tmp_path,
        compare_resnames=True,
    )


@pytest.mark.parametrize(
    "input_pdb",
    ["1FKI.cif", "9X2O.cif", "boltz2_olanzapine_P28335.cif"],
)
def test_read_cif(input_pdb, tmp_path):
    """Tests for reading and converting CIF-format structures."""
    args = "--log-level=INFO --ff=AMBER --ffout AMBER"
    output_pqr = Path(input_pdb).stem + ".pqr"
    common.run_pdb2pqr_for_tests(
        args=args,
        input_pdb=common.DATA_DIR / input_pdb,
        output_pqr=output_pqr,
        tmp_path=tmp_path,
        compare_resnames=True,
    )



"""Quatfit routines for PDB2PQR

This module is used to find the coordinates of a new atom based on a reference
set of coordinates and a definition set of coordinates.

Original Code by David J. Heisterberg, The Ohio Supercomputer Center,
1224 Kinnear Rd., Columbus, OH  43212-1163, (614)292-6036, djh@osc.edu,
djh@ohstpy.bitnet, ohstpy::djh

Translated to C from fitest.f program and interfaced with Xmol program by
Jan Labanowski, jkl@osc.edu, jkl@ohstpy.bitnet, ohstpy::jkl

.. todo::
   There are many unnecessary parameters in this module due to FORTRAN/C
   assumptions about how the code should behave.

.. codeauthor:: David Heisterberg
.. codeauthor:: Jan Labanowski
.. codeauthor:: Jens Erik Nielsen
.. codeauthor:: Todd Dolinsky
"""

import math

from .utilities import normalize


def find_coordinates(numpoints, refcoords, defcoords, defatomcoords):
    """Driver for the quaternion file.

    Provide the coordinates as inputs and obtain the coordinates for the new
    atom as output.

    :param numpoints:  the number of points in each list
    :type numpoints:  int
    :param refcoords:  the reference coordinates, a list of lists of form
        [x,y,z]
    :type refcoords:  [[float, float, float]]
    :param defcoords:  the definition coordinates, a list of lists of form
        [x,y,z]
    :type defcoords:  [[float, float, float]]
    :param defatomcoords:  the definition coordinates for the atom to be
        placed in the reference frame
    :type defatomcoords:  [[float, float, float]]
    :return: the coordinates of the new atom in the reference frame
    :rtype:  [[float, float, float]]
    """
    refcenter, fitcenter, rotation = qfit(numpoints, refcoords, defcoords)
    newcoords = qtransform(1, defatomcoords, refcenter, fitcenter, rotation)
    # Only return the first coordinates
    return newcoords[0]


def qtransform(numpoints, defcoords, refcenter, fitcenter, rotation):
    """Transform coordinates using the reference.

    Transform the set of defcoords using the reference center, the fit
    center, and a rotation matrix.

    :param numpoints:  the number of points in each list
    :type numpoints:  int
    :param defcoords:  set of coordinates to be transformed using the
        reference center and a rotation matrix
    :type defcoords:  [[float, float, float]]
    :param refcenter:  the reference center
    :type refcenter:  [[float, float, float]]
    :param fitcenter: the definition center
    :type fitcenter:  [float, float, float]
    :param rotation:  the rotation matrix
    :type rotation:  [[float, float, float]]
    :return:  the coordinates of the new point
    :rtype:  [[float, float, float]]
    """
    if numpoints == 1:
        defcoords = [defcoords]
    fitcoords = translate(numpoints, defcoords, fitcenter, 1)
    rotated = rotmol(numpoints, fitcoords, rotation)
    return translate(numpoints, rotated, refcenter, 2)


def qfit(numpoints, refcoords, defcoords):
    """Method for getting new atom coordinates from sets of reference and definition coordinates.

    .. todo::  Remove hard-coded parameters of function.

    :param numpoints:  the number of points in each list
    :type numpoints:  int
    :param refcoords:  list of reference coordinates
    :type refcoords:  [[float, float, float]]
    :param defcoords:  list of definition coordinates
    :type defcoords:  [[float, float, float]]
    :return:  (reference center, definition center, left rotation matrix)
    :rtype:  ([[float, float, float]], [[float, float, float]],
        [[float, float, float]])
    """
    nrot = 30
    refcenter, refcoords = center(numpoints, refcoords)
    defcenter, defcoords = center(numpoints, defcoords)
    _, lrot = qtrfit(numpoints, defcoords, refcoords, nrot)
    rotated = rotmol(numpoints, defcoords, lrot)
    _ = translate(numpoints, rotated, refcenter, 2)
    return refcenter, defcenter, lrot


def qchichange(initcoords, refcoords, angle):
    """Change the chiangle of the reference coordinate.

    Change the chiangle of the reference coordinate using the initcoords and
    the given angle.

    :param initcoords:  coordinates based on the point and basis atoms
        (one-dimensional list)
    :type initcoords:  [[float, float, float]]
    :param difchi:  the angle to use
    :type difchi:  float
    :param refcoords:  the atoms to analyze (list of many coordinates)
    :type refcoords:  [[float, float, float]]
    :return:  the new coordinates of the atoms
    :rtype:  [[float, float, float]]
    """
    # Initialize
    left = []
    right = []
    for _ in range(3):
        left.append(0.0)
        right.append([0.0, 0.0, 0.0])
    # Convert to radians and normalize
    radangle = math.pi * angle / 180.0
    normalized = normalize(initcoords)
    left[0] = normalized[0]
    left[1] = normalized[1]
    left[2] = normalized[2]
    # Construct the rotation matrix
    right[0][0] = math.cos(radangle) + left[0] * left[0] * (
        1.0 - math.cos(radangle)
    )
    right[1][1] = math.cos(radangle) + left[1] * left[1] * (
        1.0 - math.cos(radangle)
    )
    right[2][2] = math.cos(radangle) + left[2] * left[2] * (
        1.0 - math.cos(radangle)
    )
    right[1][0] = left[0] * left[1] * (1.0 - math.cos(radangle)) - left[
        2
    ] * math.sin(radangle)
    right[2][0] = left[0] * left[2] * (1.0 - math.cos(radangle)) + left[
        1
    ] * math.sin(radangle)
    right[0][1] = left[1] * left[0] * (1.0 - math.cos(radangle)) + left[
        2
    ] * math.sin(radangle)
    right[2][1] = left[1] * left[2] * (1.0 - math.cos(radangle)) - left[
        0
    ] * math.sin(radangle)
    right[0][2] = left[2] * left[0] * (1.0 - math.cos(radangle)) - left[
        1
    ] * math.sin(radangle)
    right[1][2] = left[2] * left[1] * (1.0 - math.cos(radangle)) + left[
        0
    ] * math.sin(radangle)
    numpoints = len(refcoords)
    return rotmol(numpoints, refcoords, right)


def rotmol(numpoints, coor, lrot):
    """Rotate a molecule

    :param numpoints:  the number of points in the list
    :type numpoints:  int
    :param coor:  the input coordinates
    :type coor:  [[float, float, float]]
    :param lrot:  the left rotation matrix
    :type lrot:  [[float, float, float]]
    :return:  the rotated coordinates
    :rtype:  [[float, float, float]]
    """
    out = []
    for i in range(numpoints):
        out.append([])
        out[i].append(
            lrot[0][0] * coor[i][0]
            + lrot[1][0] * coor[i][1]
            + lrot[2][0] * coor[i][2]
        )
        out[i].append(
            lrot[0][1] * coor[i][0]
            + lrot[1][1] * coor[i][1]
            + lrot[2][1] * coor[i][2]
        )
        out[i].append(
            lrot[0][2] * coor[i][0]
            + lrot[1][2] * coor[i][1]
            + lrot[2][2] * coor[i][2]
        )
    return out


def qtrfit(numpoints, defcoords, refcoords, nrot):
    """Find the best-fit quaternion.

    Find the quaternion, q, [and left rotation matrix, u] that minimizes

    .. math::
       | qTXq - Y | ^ 2 [|uX - Y| ^ 2]

    This is equivalent to maximizing

    .. math::
       Re(q^T X^T q Y)

    The left rotation matrix, u, is obtained from q by

    .. math::
       u = qT1q

    :param numpoints:  the number of points in each list
    :type numpoints:  int
    :param defcoords:
        list of definition coordinates, with each set a list of form [x,y,z]
    :type defcoords:  [[float, float, float]]
    :param refcoords:
        list of fitted coordinates, with each set a list of form [x,y,z]
    :type refcoords:  [[float, float, float]]
    :param nrot:  the maximum number of Jacobi sweeps
    :type nrot:  int
    :return: (the best-fit quaternion, the best-fit left rotation matrix)
    """
    xxyx = 0.0
    xxyy = 0.0
    xxyz = 0.0
    xyyx = 0.0
    xyyy = 0.0
    xyyz = 0.0
    xzyx = 0.0
    xzyy = 0.0
    xzyz = 0.0
    cmat = []
    for i in range(numpoints):
        xxyx += defcoords[i][0] * refcoords[i][0]
        xxyy += defcoords[i][0] * refcoords[i][1]
        xxyz += defcoords[i][0] * refcoords[i][2]
        xyyx += defcoords[i][1] * refcoords[i][0]
        xyyy += defcoords[i][1] * refcoords[i][1]
        xyyz += defcoords[i][1] * refcoords[i][2]
        xzyx += defcoords[i][2] * refcoords[i][0]
        xzyy += defcoords[i][2] * refcoords[i][1]
        xzyz += defcoords[i][2] * refcoords[i][2]
    for i in range(4):
        cmat.append([])
        for _ in range(4):
            cmat[i].append(0.0)
    cmat[0][0] = xxyx + xyyy + xzyz
    cmat[0][1] = xzyy - xyyz
    cmat[0][2] = xxyz - xzyx
    cmat[0][3] = xyyx - xxyy
    cmat[1][1] = xxyx - xyyy - xzyz
    cmat[1][2] = xxyy + xyyx
    cmat[1][3] = xzyx + xxyz
    cmat[2][2] = xyyy - xzyz - xxyx
    cmat[2][3] = xyyz + xzyy
    cmat[3][3] = xzyz - xxyx - xyyy
    _, vmat = jacobi(cmat, nrot)  # diagonalize c
    quat = [vmat[i][3] for i in range(4)]
    lrot = q2mat(quat)
    return quat, lrot


def jacobi(amat, nrot):
    """Jacobi diagonalizer with sorted output, only good for 4x4 matrices.

    :param amat:  Matrix to diagonalize
    :type ama:  [[float, float, float, float]]
    :param nrot:  maximum number of sweeps
    :type nrot:  int
    :return: (eigenvalues, eigenvectors)
    """
    vmat = []
    dvec = []
    the_lrot = 0
    for j in range(4):
        dvec.append(0)
        vmat.append([])
        for _ in range(4):
            vmat[j].append(0.0)
        vmat[j][j] = 1.0
        dvec[j] = amat[j][j]
    for lrot in range(nrot):
        dnorm = 0.0
        onorm = 0.0
        for j in range(4):
            dnorm += abs(dvec[j])
            for i in range(j):
                onorm += abs(amat[i][j])
        if dnorm != 0 and onorm / dnorm <= 1e-12:
            the_lrot = lrot
            break
        for j in range(1, 4):
            for i in range(j):
                bscl = amat[i][j]
                if abs(bscl) > 0.0:
                    dma = dvec[j] - dvec[i]
                    if abs(dma) + abs(bscl) <= abs(dma):
                        tscl = bscl / dma
                    else:
                        qscl = 0.5 * dma / bscl
                        tscl = 1.0 / (abs(qscl) + math.sqrt(1 + qscl * qscl))
                        if qscl < 0:
                            tscl = tscl * -1
                    cscl = 1.0 / math.sqrt(tscl * tscl + 1)
                    sscl = tscl * cscl
                    amat[i][j] = 0.0
                    for k in range(i):
                        atemp = cscl * amat[k][i] - sscl * amat[k][j]
                        amat[k][j] = sscl * amat[k][i] + cscl * amat[k][j]
                        amat[k][i] = atemp
                    for k in range(i + 1, j):
                        atemp = cscl * amat[i][k] - sscl * amat[k][j]
                        amat[k][j] = sscl * amat[i][k] + cscl * amat[k][j]
                        amat[i][k] = atemp
                    for k in range(j + 1, 4):
                        atemp = cscl * amat[i][k] - sscl * amat[j][k]
                        amat[j][k] = sscl * amat[i][k] + cscl * amat[j][k]
                        amat[i][k] = atemp
                    for k in range(4):
                        vtemp = cscl * vmat[k][i] - sscl * vmat[k][j]
                        vmat[k][j] = sscl * vmat[k][i] + cscl * vmat[k][j]
                        vmat[k][i] = vtemp
                    dtemp = (
                        cscl * cscl * dvec[i]
                        + sscl * sscl * dvec[j]
                        - 2.0 * cscl * sscl * bscl
                    )
                    dvec[j] = (
                        sscl * sscl * dvec[i]
                        + cscl * cscl * dvec[j]
                        + 2.0 * cscl * sscl * bscl
                    )
                    dvec[i] = dtemp
    nrot = the_lrot
    for j in range(3):
        k = j
        dtemp = dvec[k]
        for i in range(j + 1, 4):
            if dvec[i] < dtemp:
                k = i
                dtemp = dvec[k]
        if k > j:
            dvec[k] = dvec[j]
            dvec[j] = dtemp
            for i in range(4):
                dtemp = vmat[i][k]
                vmat[i][k] = vmat[i][j]
                vmat[i][j] = dtemp
    return dvec, vmat


def q2mat(quat):
    """Generate a left rotation matrix from a normalized quaternion

    :param quat:  the normalized quaternion
    :type quat:  [[float, float, float, float]]
    :return:  the rotation matrix
    """
    urot = []
    for i in range(3):
        urot.append([])
        for _ in range(3):
            urot[i].append(0.0)
    urot[0][0] = (
        quat[0] * quat[0]
        + quat[1] * quat[1]
        - quat[2] * quat[2]
        - quat[3] * quat[3]
    )
    urot[0][1] = 2.0 * (quat[1] * quat[2] - quat[0] * quat[3])
    urot[0][2] = 2.0 * (quat[1] * quat[3] + quat[0] * quat[2])
    urot[1][0] = 2.0 * (quat[2] * quat[1] + quat[0] * quat[3])
    urot[1][1] = (
        quat[0] * quat[0]
        - quat[1] * quat[1]
        + quat[2] * quat[2]
        - quat[3] * quat[3]
    )
    urot[1][2] = 2.0 * (quat[2] * quat[3] - quat[0] * quat[1])
    urot[2][0] = 2.0 * (quat[3] * quat[1] - quat[0] * quat[2])
    urot[2][1] = 2.0 * (quat[3] * quat[2] + quat[0] * quat[1])
    urot[2][2] = (
        quat[0] * quat[0]
        - quat[1] * quat[1]
        - quat[2] * quat[2]
        + quat[3] * quat[3]
    )
    return urot


def center(numpoints, refcoords):
    """Center a molecule using equally weighted points.

    :param numpoints:  number of points
    :type numpoints:  int
    :param refcoords:  list of reference coordinates, with each set a list of
        form [x,y,z]
    :type refcoords:  [[float, float, float]]
    :return:  (center of the set of points, moved refcoords relative to
        refcenter)
    """
    relcoords = []
    refcenter = [0.0 for _ in range(3)]
    for i in range(numpoints):
        refcenter[0] += refcoords[i][0]
        refcenter[1] += refcoords[i][1]
        refcenter[2] += refcoords[i][2]
    for i in range(3):
        refcenter[i] = refcenter[i] / numpoints
    for i in range(numpoints):
        relcoords.append([])
        relcoords[i].append(refcoords[i][0] - refcenter[0])
        relcoords[i].append(refcoords[i][1] - refcenter[1])
        relcoords[i].append(refcoords[i][2] - refcenter[2])
    return refcenter, relcoords


def translate(numpoints, refcoords, center_, mode):
    """Translate a molecule using equally weighted points.

    :param numpoints:  number of points
    :type numpoints:  int
    :param refcoords:
        list of reference coordinates, with each set a list of form [x,y,z]
    :type:  list
    :param center:  center of the system
    :type center:  [float, float, float]
    :param mode:
        if 1, center will be subtracted from refcoords; if 2, center will be
        added to refcoords
    :return:  moved refcoords relative to refcenter
    :rtype:  [[float, float, float]]
    """
    relcoords = []
    modif = 0
    if mode == 1:
        modif = -1
    elif mode == 2:
        modif = 1
    for i in range(numpoints):
        relcoords.append([])
        relcoords[i].append(refcoords[i][0] + modif * center_[0])
        relcoords[i].append(refcoords[i][1] + modif * center_[1])
        relcoords[i].append(refcoords[i][2] + modif * center_[2])
    return relcoords



"""Configuration information for PDB2PQR."""

from ._version import __version__

#: PDB2PQR version number.
VERSION = __version__


#: Options for PROPKA that PDB2PQR cannot handle
IGNORED_PROPKA_OPTIONS = {
    "thermophiles": None,
    "chains": None,
    "alignment": None,
    "mutations": None,
    "mutator": None,
    "mutator_options": None,
    "reuse_ligand_mol2_file": False,
    "keep_protons": False,
    "protonate_all": False,
}

#: How to format PDB2PQR title in output
TITLE_STR = f"PDB2PQR v{VERSION}: biomolecular structure conversion software."

#: Citation strings for PDB2PQR
CITATIONS = [
    (
        "Please cite:  Jurrus E, et al.  Improvements to the APBS "
        "biomolecular solvation software suite.  Protein Sci 27 112-128 "
        "(2018)."
    ),
    (
        "Please cite:  Dolinsky TJ, et al.  PDB2PQR: expanding and upgrading "
        "automated preparation of biomolecular structures for molecular "
        "simulations. Nucleic Acids Res 35 W522-W525 (2007)."
    ),
]

#: Standard force field names
FORCE_FIELDS = ["amber", "charmm", "parse", "tyl06", "peoepb", "swanson"]

#: Standard amino acid names
AA_NAMES = [
    "ALA",
    "ARG",
    "ASH",
    "ASN",
    "ASP",
    "CYS",
    "CYM",
    "GLN",
    "GLU",
    "GLH",
    "GLY",
    "HIS",
    "HID",
    "HIE",
    "HIP",
    "HSD",
    "HSE",
    "HSP",
    "ILE",
    "LEU",
    "LYS",
    "LYN",
    "MET",
    "PHE",
    "PRO",
    "SER",
    "THR",
    "TRP",
    "TYR",
    "TYM",
    "VAL",
]

#: Standard nucleic acid names
NA_NAMES = [
    "A",
    "A5",
    "A3",
    "C",
    "C5",
    "C3",
    "G",
    "G5",
    "G3",
    "T",
    "T5",
    "T3",
    "U",
    "U5",
    "U3",
    "RA",
    "RG",
    "RC",
    "RU",
    "DA",
    "DG",
    "DC",
    "DT",
]

#: Rename RNA residues to match PDB nomenclature (https://pdb101.rcsb.org/learn/guide-to-understanding-pdb-data/primary-sequences-and-the-pdb-format)
RNA_MAPPING = {"A": "RA", "C": "RC", "G": "RG", "U": "RU"}

#: Standard backbone atom names
BACKBONE = ["N", "CA", "C", "O", "O2", "HA", "HN", "H", "tN"]

#: A small number used by some math routines.
SMALL_NUMBER = 1.0e-7

#: Convert radians to degrees
RADIANS_TO_DEGREES = 57.2958

#: The start of warning strings to be filtered.
FILTER_WARNINGS = [
    "Skipped atom during water optimization",
    "The best donorH was not picked",
    "Multiple occupancies found",
    "Tetrahedral hydrogen reconstruction",
    "Unable to find amino or nucleic acid definition for",
]

#: Number of times warning string is printed before supressing further output
FILTER_WARNINGS_LIMIT = 10

#: Expected location for topology definition file
TOPOLOGY_DEF_PATH = "TOPOLOGY.xml"

#: Expected location for amino acid topology definition file
AA_DEF_PATH = "AA.xml"

#: Expected location for nucleic acid topology definition file
NA_DEF_PATH = "NA.xml"

#: Expected location for hydrogens topology definition file
HYD_DEF_PATH = "HYDROGENS.xml"

#: Expected location for topology patch definition file
PATCH_DEF_PATH = "PATCHES.xml"

#: Number of angle steps to scan when debumping
DEBUMP_ANGLE_STEPS = 72

#: Size of debumping step
DEBUMP_ANGLE_STEP_SIZE = float(360 // DEBUMP_ANGLE_STEPS)

#: Debump angle test
DEBUMP_ANGLE_TEST_COUNT = 10

#: Size of cells used for neighbor lookups
CELL_SIZE = 2

#: Debumping hydrogen size
BUMP_HYDROGEN_SIZE = 0.5

#: Debumping heavy atom size
BUMP_HEAVY_SIZE = 1.0

#: Disulfide bond distance limit
BONDED_SS_LIMIT = 2.5

#: Peptide bond distance limit
PEPTIDE_DIST = 1.7

#: Limit on fraction of molecule missing before giving up on repairs
REPAIR_LIMIT = 0.1

#: Cutoff for A - D - H(D) hydrogen bond angle
ANGLE_CUTOFF = 20.0

#: Cutoff for H(D) to A hydrogen bond distance
DIST_CUTOFF = 3.3

#: Avogadro's number
AVOGADRO = 6.02214076e23

#: Electron charge (in C)
ELEC_CHARGE = 1.602176634e-19

#: Vacuum permittivity (in F/m)
VACUUM_PERMIT = 8.8541878128e-12

#: Boltzmann constant (in J/K)
BOLTZMANN = 1.380649e-23

#: Charge deviation considered to be non-integer
CHARGE_ERROR = 1e-3



"""Routines for biomolecule optimization.

.. codeauthor::  Jens Erik Nielsen
.. codeauthor::  Todd Dolinsky
.. codeauthor::  Yong Huang
.. codeauthor::  Nathan Baker
"""

import logging

from . import aa, cells, io
from . import quatfit as quat
from . import utilities as util
from .config import (
    BUMP_HEAVY_SIZE,
    BUMP_HYDROGEN_SIZE,
    CELL_SIZE,
    DEBUMP_ANGLE_STEP_SIZE,
    DEBUMP_ANGLE_STEPS,
    DEBUMP_ANGLE_TEST_COUNT,
    SMALL_NUMBER,
)

_LOGGER = logging.getLogger(__name__)
_LOGGER.addFilter(io.DuplicateFilter())


class Debump:
    """Grab bag of random stuff that apparently didn't fit elsewhere.

    .. todo::
        This class needs to be susbtantially refactored in to multiple classes
        with clear responsibilities.
    """

    def __init__(self, biomolecule, definition=None):
        """Initialize the Debump object.

        :param biomolecule:  the biomolecule to debump
        :type biomolecule:  Biomolecule
        :param definition:  topology definition file
        :type definition:  Definition
        """
        self.biomolecule = biomolecule
        self.definition = definition
        self.aadef = None
        self.cells = {}
        if definition is not None:
            self.aadef = definition.getAA()
            self.nadef = definition.getNA()

    def get_bump_score(self, residue):
        """Get a bump score for a residue.

        :param residue:  residue with bumping to evaluate
        :type residue:  Residue
        :return:  bump score
        :rtype:  float
        """
        # Do some setup
        self.cells = cells.Cells(CELL_SIZE)
        self.cells.assign_cells(self.biomolecule)
        self.biomolecule.calculate_dihedral_angles()
        self.biomolecule.set_donors_acceptors()
        self.biomolecule.update_internal_bonds()
        self.biomolecule.set_reference_distance()
        bumpscore = 0.0
        if not isinstance(residue, aa.Amino):
            return 0.0
        # Initialize variables
        for atom in residue.atoms:
            atomname = atom.name
            if atomname[0] != "H":
                continue
            bumpscore += self.get_bump_score_atom(atom)
        return bumpscore

    def get_bump_score_atom(self, atom):
        """Find nearby atoms for conflict-checking.

        Uses neighboring cells to compare atoms rather than an all versus all
        O(n^2) algorithm, which saves a great deal of time.  There are several
        instances where we ignore potential conflicts; these include
        donor/acceptor pairs, atoms in the same residue, and bonded CYS
        bridges.

        :param atom:  find nearby atoms to this atom
        :type atom:  Atom
        :return:  a bump score sum((dist-cutoff)**20 for all nearby atoms
        :rtype:  float
        """
        # Initialize some variables
        residue = atom.residue
        atom_size = BUMP_HYDROGEN_SIZE if atom.is_hydrogen else BUMP_HEAVY_SIZE
        # Get atoms from nearby cells
        closeatoms = self.cells.get_near_cells(atom)
        # Loop through and see if any are within the cutoff
        bumpscore = 0.0
        for closeatom in closeatoms:
            closeresidue = closeatom.residue
            if closeresidue == residue and (
                closeatom in atom.bonds or atom in closeatom.bonds
            ):
                continue
            if not isinstance(closeresidue, aa.Amino):
                continue
            if (
                isinstance(residue, aa.CYS)
                and residue.ss_bonded_partner == closeatom
            ):
                continue
            if (
                atom.is_hydrogen
                and len(atom.bonds) != 0
                and atom.bonds[0].hdonor
                and closeatom.hacceptor
            ):
                continue
            if (
                closeatom.is_hydrogen
                and len(closeatom.bonds) != 0
                and closeatom.bonds[0].hdonor
                and atom.hacceptor
            ):
                continue
            dist = util.distance(atom.coords, closeatom.coords)
            other_size = (
                BUMP_HYDROGEN_SIZE
                if closeatom.is_hydrogen
                else BUMP_HEAVY_SIZE
            )
            cutoff = atom_size + other_size
            if dist < cutoff:
                bumpscore += 1000.0
        _LOGGER.debug(f"BUMPSCORE {bumpscore!s}")
        return bumpscore

    def debump_biomolecule(self):
        """Minimize bump score for molecule.

        Make sure that none of the added atoms were rebuilt on top of existing
        atoms. See each called function for more information.

        :raises ValueError:  if missing (backbone) atoms are encountered
        """
        # Do some setup
        self.cells = cells.Cells(CELL_SIZE)
        self.cells.assign_cells(self.biomolecule)
        self.biomolecule.calculate_dihedral_angles()
        self.biomolecule.set_donors_acceptors()
        self.biomolecule.update_internal_bonds()
        try:
            self.biomolecule.set_reference_distance()
        except ValueError as err:
            err = f"Biomolecular structure is incomplete:  {err}"
            raise ValueError(err)
        # Determine which residues to debump
        for residue in self.biomolecule.residues:
            if not isinstance(residue, aa.Amino):
                continue
            # Initialize variables
            conflict_names = self.find_residue_conflicts(
                residue, write_conflict_info=True
            )
            if not conflict_names:
                continue
            # Otherwise debump the residue
            _LOGGER.debug(f"Starting to debump {residue}...")
            _LOGGER.debug(
                f"Debumping cutoffs: {BUMP_HEAVY_SIZE * 2:2.1f} for "
                f"heavy-heavy, {BUMP_HYDROGEN_SIZE + BUMP_HEAVY_SIZE:2.1f} "
                f"for hydrogen-heavy, and {BUMP_HYDROGEN_SIZE * 2:2.1f} "
                f"for hydrogen-hydrogen."
            )
            if self.debump_residue(residue, conflict_names):
                _LOGGER.debug("Debumping Successful!")
            else:
                text = f"WARNING: Unable to debump {residue}"
                _LOGGER.warning(text)
        _LOGGER.debug("Done checking if we must debump any residues.")

    def find_residue_conflicts(self, residue, *, write_conflict_info=False):
        """Find conflicts between residues.

        :param residue:  residue to check
        :type residue:  Residue
        :param write_conflict_info:  write verbose output about conflict
        :type write_conflict_info:  bool
        :return: list of conflicts
        :rtype: [str]
        """
        conflict_names = []
        for atom in residue.atoms:
            atomname = atom.name
            if not atom.added:
                continue
            if atomname == "H":
                continue
            if atom.optimizeable:
                continue
            nearatoms = self.find_nearby_atoms(atom)
            # If something is too close, we must debump the residue
            if nearatoms != {}:
                conflict_names.append(atomname)
                if write_conflict_info:
                    for repatom in nearatoms:
                        _LOGGER.debug(
                            f"{residue} {atomname} is too close to "
                            f"{repatom.residue} {repatom.name}"
                        )
        return conflict_names

    def score_dihedral_angle(self, residue, anglenum):
        """Assign score to dihedral angle.

        :param residue:  residue with dihedral angles to score
        :type residue:  Residue
        :param anglenum:  specific dihedral angle index
        :type anglenum:  int
        :return:  score for dihedral angle
        :rtype:  float
        """
        score = 0
        atomnames = residue.reference.dihedrals[anglenum].split()
        pivot = atomnames[2]
        moveablenames = residue.get_moveable_names(pivot)
        for name in moveablenames:
            nearatoms = self.find_nearby_atoms(residue.get_atom(name))
            for value in nearatoms.values():
                score += value
        return score

    def debump_residue(self, residue, conflict_names):
        """Debump a specific residue.

        Only should be called if the residue has been detected to have a
        conflict.
        If called, try to rotate about dihedral angles to resolve the conflict.

        :param residue:  the residue in question
        :type residue:  Residue
        :param conflict_names:  a list of atomnames that were rebuilt too close
            to other atoms
        :type conflict_names:  [str]
        :return: True if successful, False otherwise
        :rtype: bool
        """
        # Initialize some variables
        anglenum = -1
        curr_conflict_names = conflict_names
        # Try to find a workable solution
        for _ in range(DEBUMP_ANGLE_TEST_COUNT):
            anglenum = residue.pick_dihedral_angle(
                curr_conflict_names, anglenum
            )
            if anglenum == -1:
                return False
            _LOGGER.debug(
                f"Using dihedral angle number {anglenum} to debump "
                "the residue."
            )
            bestscore = self.score_dihedral_angle(residue, anglenum)
            found_improved = False
            bestangle = orig_angle = residue.dihedrals[anglenum]
            # Skip the first angle as it's already known.
            for i in range(1, DEBUMP_ANGLE_STEPS):
                newangle = orig_angle + (DEBUMP_ANGLE_STEP_SIZE * i)
                self.set_dihedral_angle(residue, anglenum, newangle)
                # Check for conflicts
                score = self.score_dihedral_angle(residue, anglenum)
                if score == 0:
                    if not self.find_residue_conflicts(residue):
                        _LOGGER.debug(
                            f"No conflicts found at angle {newangle!r}"
                        )
                        return True
                    else:
                        bestangle = newangle
                        found_improved = True
                        break
                # Set the best angle
                elif score < bestscore:
                    diff = abs(bestscore - score)
                    # Don't update if it's effectively a tie
                    if diff > SMALL_NUMBER:
                        bestscore = score
                        bestangle = newangle
                        found_improved = True
            self.set_dihedral_angle(residue, anglenum, bestangle)
            curr_conflict_names = self.find_residue_conflicts(residue)
            if found_improved:
                err = f"Best score of {bestscore} at angle {bestangle}."
                _LOGGER.debug(err)
                _LOGGER.debug(f"New conflict set: {curr_conflict_names}")
            else:
                _LOGGER.debug("No improvement found for this dihedral angle.")
        # If we're here, debumping was unsuccessful
        return False

    def get_closest_atom(self, atom):
        """Get the closest atom that does not form a donor/acceptor pair.

        Used to detect potential conflicts.

        .. note::  Cells must be set before using this function.

        :param atom:  the atom to test
        :type atom:  Atom
        :return:  the closest atom to the input atom that does not satisfy a
            donor/acceptor pair.
        :rtype:  Atom
        """
        # Initialize some variables
        bestdist = 999.99
        bestwatdist = 999.99
        bestatom = None
        bestwatatom = None
        residue = atom.residue
        # Get atoms from nearby cells
        closeatoms = self.cells.get_near_cells(atom)
        # Loop through and see which is the closest
        for closeatom in closeatoms:
            closeresidue = closeatom.residue
            if closeresidue == residue:
                continue
            if not isinstance(closeresidue, (aa.Amino, aa.WAT)):
                continue
            if (
                isinstance(residue, aa.CYS)
                and residue.ss_bonded_partner == closeatom
            ):
                continue
            # Also ignore if this is a donor/acceptor pair
            if (
                atom.is_hydrogen
                and atom.bonds[0].hdonor
                and closeatom.hacceptor
            ):
                continue
            if (
                closeatom.is_hydrogen
                and closeatom.bonds[0].hdonor
                and atom.hacceptor
            ):
                continue
            dist = util.distance(atom.coords, closeatom.coords)
            if isinstance(closeresidue, aa.WAT):
                if dist < bestwatdist:
                    bestwatdist = dist
                    bestwatatom = closeatom
            else:
                if dist < bestdist:
                    bestdist = dist
                    bestatom = closeatom
        if bestdist > bestwatdist:
            txt = (
                f"Skipped atom during water optimization: {bestwatatom.name} "
                f"in {bestwatatom.residue} skipped "
                f"when optimizing {atom.name} in {residue}"
            )
            _LOGGER.warning(txt)
        return bestatom

    def find_nearby_atoms(self, atom):
        """Find nearby atoms for conflict-checking.

        Uses neighboring cells to compare atoms rather than an all versus all
        O(n^2) algorithm, which saves a great deal of time.
        There are several instances where we ignore potential conflicts; these
        include donor/acceptor pairs, atoms in the same residue, and bonded CYS
        bridges.

        :param atom:  find nearby atoms to this atom
        :type atom:  Atom
        :return:  a dictionary of ``Atom too close`` to ``amount of overlap
            for that atom``
        :rtype:  dict
        """
        # Initialize some variables
        nearatoms = {}
        residue = atom.residue
        atom_size = BUMP_HYDROGEN_SIZE if atom.is_hydrogen else BUMP_HEAVY_SIZE
        # Get atoms from nearby cells
        closeatoms = self.cells.get_near_cells(atom)
        # Loop through and see if any are within the cutoff
        for closeatom in closeatoms:
            closeresidue = closeatom.residue
            if closeresidue == residue and (
                closeatom in atom.bonds or atom in closeatom.bonds
            ):
                continue
            if not isinstance(closeresidue, (aa.Amino, aa.WAT)):
                continue
            if (
                isinstance(residue, aa.CYS)
                and residue.ss_bonded_partner == closeatom
            ):
                continue
            # Also ignore if this is a donor/acceptor pair
            if (
                atom.is_hydrogen
                and len(atom.bonds) != 0
                and atom.bonds[0].hdonor
                and closeatom.hacceptor
            ):
                continue
            if (
                closeatom.is_hydrogen
                and len(closeatom.bonds) != 0
                and closeatom.bonds[0].hdonor
                and atom.hacceptor
            ):
                continue
            dist = util.distance(atom.coords, closeatom.coords)
            other_size = (
                BUMP_HYDROGEN_SIZE
                if closeatom.is_hydrogen
                else BUMP_HEAVY_SIZE
            )
            cutoff = atom_size + other_size
            if dist < cutoff:
                nearatoms[closeatom] = cutoff - dist
        return nearatoms

    def set_dihedral_angle(self, residue, anglenum, angle):
        """Rotate a residue about a given angle.

        Uses :mod:`quatfit` methods to perform the matrix mathematics.

        :param residue:  residue to rotate
        :type residue:  Residue
        :param anglenum:  specific dihedral angle index
        :param angle:  the angle to set the dihedral to
        :type angle:  float
        """
        coordlist = []
        initcoords = []
        movecoords = []
        pivot = ""
        oldangle = residue.dihedrals[anglenum]
        diff = angle - oldangle
        atomnames = residue.reference.dihedrals[anglenum].split()
        pivot = atomnames[2]
        for atomname in atomnames:
            if residue.has_atom(atomname):
                coordlist.append(residue.get_atom(atomname).coords)
            else:
                raise ValueError("Error occurred while trying to debump!")
        initcoords = util.subtract(coordlist[2], coordlist[1])
        moveablenames = residue.get_moveable_names(pivot)
        for name in moveablenames:
            atom = residue.get_atom(name)
            movecoords.append(util.subtract(atom.coords, coordlist[1]))
        newcoords = quat.qchichange(initcoords, movecoords, diff)
        for iatom, atom_name in enumerate(moveablenames):
            atom = residue.get_atom(atom_name)
            self.cells.remove_cell(atom)
            atom.x = newcoords[iatom][0] + coordlist[1][0]
            atom.y = newcoords[iatom][1] + coordlist[1][1]
            atom.z = newcoords[iatom][2] + coordlist[1][2]
            self.cells.add_cell(atom)
        # Set the new angle
        coordlist = []
        for atomname in atomnames:
            if residue.has_atom(atomname):
                coordlist.append(residue.get_atom(atomname).coords)
            else:
                raise ValueError("Error occurred while trying to debump!")
        dihed = util.dihedral(
            coordlist[0], coordlist[1], coordlist[2], coordlist[3]
        )
        residue.dihedrals[anglenum] = dihed



"""PDB2PQR

This package takes a PDB file as input and performs optimizations before
yielding a new PDB-style file as output.

For more information, see http://www.poissonboltzmann.org/
"""

import logging

from ._version import __version__
from .main import main as pdb2pqr_main
from .main import run_pdb2pqr

_LOGGER = logging.getLogger(__name__)
logging.captureWarnings(capture=True)


if __name__ == "__main__":
    pdb2pqr_main()



from pathlib import Path

from pdb2pqr.definitions import Definition


def test_dt5_patch_applies_to_dt_reference() -> None:
    dat_dir = Path(__file__).resolve().parents[1] / "pdb2pqr" / "dat"

    with (
        (dat_dir / "AA.xml").open() as aa_file,
        (dat_dir / "NA.xml").open() as na_file,
        (dat_dir / "PATCHES.xml").open() as patch_file,
    ):
        definition = Definition(aa_file, na_file, patch_file)

    assert "DT" in definition.map
    assert "DT5" in definition.map
    assert "P" not in definition.map["DT5"].map
    assert "O1P" not in definition.map["DT5"].map
    assert "H5T" in definition.map["DT5"].map



"""Tests of I/O functions."""

import logging
from difflib import Differ
from pathlib import Path

import pytest

from pdb2pqr.io import read_dx, read_pqr, read_qcd, write_cube

_LOGGER = logging.getLogger(__name__)
DATA_DIR = Path("tests/data")
PQR_LIST = list(DATA_DIR.glob("**/*.pqr"))


@pytest.mark.parametrize("input_pqr", PQR_LIST, ids=str)
def test_read_pqr(input_pqr):
    """Test that :func:`read_pqr` doesn't raise an error.

    Doesn't test functionality since that is implicit in several other tests
    that parse generated PQR output.

    :param input_pqr:  path to PQR file to test
    :type input_pqr:  str
    """
    with open(input_pqr) as pqr_file:
        read_pqr(pqr_file)


def test_read_qcd():
    """Test that :func:`read_pqr` doesn't raise an error.

    Doesn't test functionality.
    """
    qcd_path = DATA_DIR / "dummy.qcd"
    with open(qcd_path) as qcd_file:
        read_qcd(qcd_file)


def test_dx2cube(tmp_path):
    """Test conversion of OpenDX files to Cube files."""
    pqr_path = DATA_DIR / "dx2cube.pqr"
    dx_path = DATA_DIR / "dx2cube.dx"
    cube_gen = tmp_path / "test.cube"
    cube_test = DATA_DIR / "dx2cube.cube"
    _LOGGER.info(f"Reading PQR from {pqr_path}...")
    with open(pqr_path) as pqr_file:
        atom_list = read_pqr(pqr_file)
    _LOGGER.info(f"Reading DX from {dx_path}...")
    with open(dx_path) as dx_file:
        dx_dict = read_dx(dx_file)
    _LOGGER.info(f"Writing Cube to {cube_gen}...")
    with open(cube_gen, "w") as cube_file:
        write_cube(cube_file, dx_dict, atom_list)
    _LOGGER.info(f"Reading this cube from {cube_gen}...")
    this_lines = [line.strip() for line in open(cube_gen)]
    _LOGGER.info(f"Reading test cube from {cube_test}...")
    test_lines = [line.strip() for line in open(cube_test)]
    differ = Differ()
    differences = [
        line
        for line in differ.compare(this_lines, test_lines)
        if line[0] != " "
    ]

    if differences:
        for diff in differences:
            _LOGGER.error(f"Found difference:  {diff}")
        raise ValueError
    _LOGGER.info("No differences found in output")



"""Nucleic Acid Structures for PDB2PQR

This module contains the base nucleic acid structures for pdb2pqr.

.. codeauthor::  Todd Dolinsky
.. codeauthor::  Nathan Baker
"""

from . import residue
from . import structures as struct
from .definitions import DefinitionResidue
from .pdb import ATOM, HETATM


class Nucleic(residue.Residue):
    """This class provides standard features of the nucleic acids listed below."""

    def __init__(self, atoms: list[ATOM | HETATM], ref: DefinitionResidue):
        sample_atom = atoms[-1]

        self.atoms: list[struct.Atom] = []
        self.name = sample_atom.res_name
        self.chain_id = sample_atom.chain_id
        self.res_seq = sample_atom.res_seq
        self.ins_code = sample_atom.ins_code

        self.ffname = self.name
        self.map: dict[str, struct.Atom] = {}
        self.dihedrals = []
        self.patches = []
        self.is3term = 0
        self.is5term = 0
        self.is_c_term = 0
        self.is_n_term = 0
        self.missing = []
        self.reference = ref

        # Create each atom
        for atom in atoms:
            if atom.name in ref.altnames:  # Rename atoms
                atom.name = ref.altnames[atom.name]

            if atom.name not in self.map:
                atom_ = struct.Atom(atom, "ATOM", self)
                self.add_atom(atom_)

    def create_atom(
        self,
        atomname: str,
        newcoords: list[float] | tuple[float, float, float],
    ):
        """Create an atom.

        Overrides the generic residue's create_atom().

        .. todo:: This code is duplicated in several places.

        :param atomname:  the name of the atom to add
        :type atomname:  str
        :param newcoords:  the coordinates of the atom
        :type newcoords:  [(float, float, float)]
        """
        oldatom = self.atoms[0]
        newatom = struct.Atom(oldatom, "ATOM", self)
        newatom.x = newcoords[0]
        newatom.y = newcoords[1]
        newatom.z = newcoords[2]
        newatom.name = atomname
        newatom.element = atomname[0]
        newatom.occupancy = 1.0
        newatom.temp_factor = 0.0
        newatom.added = 1
        self.add_atom(newatom)

    def add_atom(self, atom: struct.Atom):
        """Add existing atom to system.

        Override the existing add_atom - include the link to the reference
        object.

        :param atom:  atom to add to system.
        :type atom:  Atom
        """
        self.atoms.append(atom)
        atomname = atom.name
        self.map[atomname] = atom
        try:
            atom.reference = self.reference.map[atomname]
            for bond in atom.reference.bonds:
                if self.has_atom(bond):
                    bondatom = self.map[bond]
                    if bondatom not in atom.bonds:
                        atom.bonds.append(bondatom)
                    if atom not in bondatom.bonds:
                        bondatom.bonds.append(atom)
        except KeyError:
            atom.reference = None

    def add_dihedral_angle(self, value):
        """Add the value to the list of chi angles.

        :param value:  dihedral angle to add to list (in degrees)
        :type value:  float
        """
        self.dihedrals.append(value)

    def set_state(self):
        """Adds the termini for all inherited objects."""
        if self.is5term:
            self.ffname = self.ffname + "5"
        if self.is3term:
            self.ffname = self.ffname + "3"


class ADE(Nucleic):
    """Adenosine class."""

    def __init__(self, atoms, ref: DefinitionResidue):
        """Initialize residue.

        :param atoms:  add atoms to residue
        :type atoms:  [Atom]
        """
        Nucleic.__init__(self, atoms, ref)
        self.reference = ref

    def letter_code(self):
        """Return letter code for nucleic acid.

        :return:  letter code for nucleic acid
        :rtype:  str
        """
        return "A"

    def set_state(self):
        """Set ribo- vs. deoxyribo- state of this residue."""
        self.ffname = "RA" if self.has_atom("O2'") else "DA"
        Nucleic.set_state(self)


DA = ADE
RA = ADE
A = ADE


class CYT(Nucleic):
    """Cytidine class"""

    def __init__(self, atoms, ref: DefinitionResidue):
        """Initialize residue.

        :param atoms:  add atoms to residue
        :type atoms:  [Atom]
        """
        Nucleic.__init__(self, atoms, ref)
        self.reference = ref

    def letter_code(self):
        """Return letter code for nucleic acid.

        :return:  letter code for nucleic acid
        :rtype:  str
        """
        return "C"

    def set_state(self):
        """Set ribo- vs. deoxyribo- state of this residue."""
        self.ffname = "RC" if self.has_atom("O2'") else "DC"
        Nucleic.set_state(self)


DC = CYT
RC = CYT
C = CYT


class GUA(Nucleic):
    """Guanosine class"""

    def __init__(self, atoms, ref: DefinitionResidue):
        """Initialize residue.

        :param atoms:  add atoms to residue
        :type atoms:  [Atom]
        """
        Nucleic.__init__(self, atoms, ref)
        self.reference = ref

    def letter_code(self):
        """Return letter code for nucleic acid.

        :return:  letter code for nucleic acid
        :rtype:  str
        """
        return "G"

    def set_state(self):
        """Set ribo- vs. deoxyribo- state of this residue."""
        self.ffname = "RG" if self.has_atom("O2'") else "DG"
        Nucleic.set_state(self)


DG = GUA
RG = GUA
G = GUA


class THY(Nucleic):
    """Thymine class"""

    def __init__(self, atoms, ref: DefinitionResidue):
        """Initialize residue.

        :param atoms:  add atoms to residue
        :type atoms:  [Atom]
        """
        Nucleic.__init__(self, atoms, ref)
        self.reference = ref

    def letter_code(self):
        """Return letter code for nucleic acid.

        :return:  letter code for nucleic acid
        :rtype:  str
        """
        return "T"

    def set_state(self):
        """Set ribo- vs. deoxyribo- state of this residue."""
        self.ffname = "DT"
        Nucleic.set_state(self)


DT = THY


class URA(Nucleic):
    """Uridine class"""

    def __init__(self, atoms, ref: DefinitionResidue):
        """Initialize residue.

        :param atoms:  add atoms to residue
        :type atoms:  [Atom]
        """
        Nucleic.__init__(self, atoms, ref)
        self.reference = ref

    def letter_code(self):
        """Return letter code for nucleic acid.

        :return:  letter code for nucleic acid
        :rtype:  str
        """
        return "U"

    def set_state(self):
        """Set ribo- vs. deoxyribo- state of this residue."""
        self.ffname = "RU"
        Nucleic.set_state(self)


RU = URA
U = URA



"""Topology-related classes for hydrogen optimization."""

import logging
from xml import sax

from .. import aa
from .. import definitions as defns
from .. import utilities as util
from ..config import ANGLE_CUTOFF, DIST_CUTOFF
from . import io, optimize

_LOGGER = logging.getLogger(__name__)
_LOGGER.addFilter(io.DuplicateFilter())


class Flip(optimize.Optimize):
    """The holder for optimization of flippable residues."""

    def __init__(self, residue, optinstance, routines):
        """Initialize a potential flip.

        Rather than flipping the given residue back and forth, take each atom
        that would be flipped and pre-flip it, making a new ``*FLIP`` atom in
        its place.

        :param residue:  the residue to flip
        :type residue:  Residue
        :param optinstance:  the optimization instance containing information
            about what to optimize
        :type optinstance:  OptimizationHandler
        :param routines:  debumping routines object
        :type routines:  Debump
        """
        # Initialize some variables
        self.optinstance = optinstance
        self.residue = residue
        self.routines = routines
        self.atomlist = []
        self.hbonds = []
        map_ = {}
        # Get all moveable names for this angle/residue pair
        dihedral = optinstance.optangle
        pivot = dihedral.split()[2]
        moveablenames = residue.get_moveable_names(pivot)
        # HO in CTERM shouldn't be in the list of flip atoms
        if residue.is_c_term:
            newmoveablenames = [name for name in moveablenames if name != "HO"]
            moveablenames = newmoveablenames
        # Cache current coordinates
        for name in moveablenames:
            atom = residue.get_atom(name)
            map_[name] = atom.coords
        # Flip the residue about the angle
        anglenum = residue.reference.dihedrals.index(dihedral)
        if anglenum == -1:
            raise ValueError("Unable to find Flip dihedral angle!")
        newangle = 180.0 + residue.dihedrals[anglenum]
        self.routines.set_dihedral_angle(residue, anglenum, newangle)
        # Create new atoms at cached positions
        for name, value in map_.items():
            newname = f"{name}FLIP"
            residue.create_atom(newname, value)
            newatom = residue.get_atom(newname)
            self.routines.cells.add_cell(newatom)
            # Set the bonds
            newatom.reference = residue.reference.map[name]
            for bond in newatom.reference.bonds:
                newbond = f"{bond}FLIP"
                if residue.has_atom(newbond):
                    bondatom = residue.map[newbond]
                    if bondatom not in newatom.bonds:
                        newatom.bonds.append(bondatom)
                    if newatom not in bondatom.bonds:
                        bondatom.bonds.append(newatom)
                # And connect back to the existing structure
                newbond = bond
                if residue.has_atom(newbond):
                    bondatom = residue.map[newbond]
                    if bondatom not in newatom.bonds:
                        newatom.bonds.append(bondatom)
                    if newatom not in bondatom.bonds:
                        bondatom.bonds.append(newatom)
        residue.set_donors_acceptors()
        # Add to the optimization list
        for name in moveablenames:
            # Get the atom
            atom = residue.get_atom(name)
            if not atom.is_hydrogen and (atom.hdonor or atom.hacceptor):
                self.atomlist.append(atom)
            # And the FLIP
            atom = residue.get_atom(f"{name}FLIP")
            if not atom.is_hydrogen and (atom.hdonor or atom.hacceptor):
                self.atomlist.append(atom)
        # Special case: Neutral unassigned HIS can be acceptors
        if (
            isinstance(residue, aa.HIS)
            and residue.name == "HIS"
            and len(residue.patches) == 1
        ):
            for atom in self.atomlist:
                if atom.name.startswith("N"):
                    atom.hacceptor = 1

    def try_both(self, donor, acc, accobj):
        """Try to optimize both donor and acceptor bonds.

        Called when both the donor and acceptor are optimizeable.
        If one is fixed, we only need to try one side.
        Otherwise first try to satisfy the donor - if that's succesful, try to
        satisfy the acceptor.
        An undo may be necessary if the donor is satisfied and the acceptor
        isn't.

        :param donor:  hydrogen bond donor
        :type donor:  Atom
        :param acc:  hydrogen bond acceptor
        :type acc:  Atom
        :param accobj:  a Flip-like hydrogen bond structure object
        :type accobj:  Flip
        :return:  indication of whether hydrogen was added
        :rtype:  bool
        """
        # If one residue if fixed, use other functions
        if donor.residue.fixed:
            return bool(accobj.try_acceptor(acc, donor))
        if acc.residue.fixed:
            return bool(self.try_donor(donor, acc))
        _LOGGER.debug(
            f"Working on {donor.residue} {donor.name} (donor) "
            f"to {acc.residue} {acc.name} (acceptor)"
        )
        if self.is_hbond(donor, acc):
            if accobj.try_acceptor(acc, donor):
                self.fix_flip(donor)
                donor.hacceptor = 0
                _LOGGER.debug("NET BOND SUCCESSFUL!")
                return True
            return False
        return False

    def try_donor(self, donor, acc):
        """Add hydrogen to an optimizable residue.

        :param donor:  hydrogen bond donor
        :type donor:  Atom
        :param acc:  hydrogen bond acceptor
        :type acc:  Atom
        :return:  indication of whether addition was successful
        :rtype:  bool
        """
        residue = self.residue
        # Do some error checking
        if not acc.hacceptor:
            return False
        _LOGGER.debug(
            f"Working on {donor.residue} {donor.name} (donor) "
            f"to {acc.residue} {acc.name} (acceptor)"
        )
        if self.is_hbond(donor, acc):
            residue.fixed = donor.name
            self.fix_flip(donor)
            donor.hacceptor = 0
            return True
        return False

    def try_acceptor(self, acc, donor):
        """Aa lone pair to an optimizable residue.

        :param acc:  hydrogen bond acceptor
        :type acc:  Atom
        :param donor:  hydrogen bond donor
        :type donor:  Atom
        :return:  indication of whether addition was successful
        :rtype:  bool
        """
        residue = acc.residue
        # Do some error checking
        if not donor.hdonor:
            return False
        _LOGGER.debug(
            f"Working on {acc.residue} {acc.name} (acceptor) "
            f"to {donor.residue} {donor.name} (donor)"
        )
        if self.is_hbond(donor, acc):
            residue.fixed = acc.name
            self.fix_flip(acc)
            acc.hdonor = 0
            return True
        return False

    def fix_flip(self, bondatom):
        """Remove atoms if hydrogen bond has been found for specified atom.

        If bondatom is of type ``*FLIP``, remove all ``*`` atoms, otherwise
        remove all ``*FLIP`` atoms.

        :param bondatom:  atom to flip
        :type bondatom:  Atom
        """
        residue = bondatom.residue
        # Initialize some variables
        atomlist = list(residue.atoms)
        # Set a flag to see whether to delete the FLIPs or not
        flag = 0
        if bondatom.name.endswith("FLIP"):
            flag = 1
        dstr = f"fix_flip called for residue {residue}, bondatom {bondatom} "
        dstr += f"and flag {flag:d}"
        _LOGGER.debug(dstr)
        residue.wasFlipped = flag == 0
        # Delete the appropriate atoms
        for atom in atomlist:
            atomname = atom.name
            if atomname.endswith("FLIP") and flag:  # Delete the other list
                if residue.has_atom(atomname[:-4]):
                    self.routines.cells.remove_cell(
                        residue.get_atom(atomname[:-4])
                    )
                    residue.remove_atom(atomname[:-4])
            elif atomname.endswith("FLIP"):  # Delete the flip
                self.routines.cells.remove_cell(atom)
                residue.remove_atom(atomname)
            else:
                continue
        residue.fixed = 1

    def finalize(self):
        """Finalize a flippable group back to its original state.

        Since the original atoms are now ``*FLIP``, it deletes the ``*`` atoms
        and renames the ``*FLIP`` atoms back to ``*``.
        """
        residue = self.residue
        if residue.fixed:
            return
        atomlist = list(residue.atoms)
        for atom in atomlist:
            if atom.name.endswith("FLIP"):
                self.routines.cells.remove_cell(atom)
                residue.remove_atom(atom.name[:-4])
                residue.rename_atom(atom.name, atom.name[:-4])
        residue.fixed = 1

    def complete(self):
        """Complete the flippable residue optimization.

        Call the :func:`finalize` function, and then rename all ``FLIP`` atoms
        back to their standard names.
        """
        residue = self.residue
        self.finalize()
        # Rename all *FLIP atoms
        for atom in residue.atoms:
            atomname = atom.name
            if atomname.endswith("FLIP"):
                residue.rename_atom(atomname, atomname[:-4])


class Generic(optimize.Optimize):
    """Generic optimization class"""

    def __init__(self, residue, optinstance, routines):
        """Initialize a potential optimization.

        :param residue:  the residue to optimize
        :type residue:  Residue
        :param optinstance:  the optimization instance containing information
            about what to optimize
        :type optinstance:  OptimizationHandler
        :param routines:  debumping routines object
        :type routines:  Debump
        """
        self.optinstance = optinstance
        self.residue = residue
        self.routines = routines
        self.hbonds = []
        self.atomlist = []

    def finalize(self):
        """Initialize some variable and pass."""
        pass

    def complete(self):
        """If not already fixed, finalize."""
        if not self.residue.fixed:
            self.finalize()


class Alcoholic(optimize.Optimize):
    """The class for alcoholic residue."""

    def __init__(self, residue, optinstance, routines):
        """Initialize the alcoholic class by removing the alcoholic hydrogen if it exists.

        :param residue:  the residue to optimize
        :type residue:  Residue
        :param optinstance:  the optimization instance containing information
            about what to optimize
        :type optinstance:  OptimizationHandler
        :param routines:  debumping routines object
        :type routines:  Debump
        """
        # Initialize some variables
        self.optinstance = optinstance
        self.residue = residue
        self.routines = routines
        self.atomlist = []
        self.hbonds = []
        name = next(iter(optinstance.map.keys()))
        self.hname = name
        bondname = residue.reference.get_atom(name).bonds[0]
        self.atomlist.append(residue.get_atom(bondname))
        if residue.has_atom(name):
            atom = residue.get_atom(name)
            self.routines.cells.remove_cell(atom)
            residue.remove_atom(name)

    def try_both(self, donor, acc, accobj):
        """Attempt to optimize both the donor and acceptor.

        If one is fixed, we only need to try one side.
        Otherwise first try to satisfy the donor - if that's succesful, try
        to satisfy the acceptor.  An undo may be necessary if the donor is
        satisfied and the acceptor isn't.

        :param donor:  hydrogen bond donor
        :type donor:  Atom
        :param acc:  hydrogen bond acceptor
        :type acc:  Atom
        :param accobj:  a Flip-like hydrogen bond structure object
        :type accobj:  Flip
        :return:  indication of whether hydrogen was added
        :rtype:  bool
        """
        # If one residue if fixed, use other functions
        residue = donor.residue
        if donor.residue.fixed:
            return bool(accobj.try_acceptor(acc, donor))
        if acc.residue.fixed:
            return bool(self.try_donor(donor, acc))
        if self.try_donor(donor, acc):
            if accobj.try_acceptor(acc, donor):
                _LOGGER.debug("NET BOND SUCCESSFUL!")
                return True
            residue.remove_atom(self.hname)
            _LOGGER.debug("REMOVED NET HBOND")
            return False
        return False

    def try_donor(self, donor, acc):
        """Add hydrogen to an optimizable residue.

        :param donor:  hydrogen bond donor
        :type donor:  Atom
        :param acc:  hydrogen bond acceptor
        :type acc:  Atom
        :return:  indication of whether addition was successful
        :rtype:  bool
        """
        residue = self.residue
        # Do some error checking
        if not acc.hacceptor:
            return False
        # Get the name of the atom to add
        newname = self.hname
        if residue.has_atom(newname):
            return False
        _LOGGER.debug(
            f"Working on {donor.residue} {donor.name} (donor) "
            f"to {acc.residue} {acc.name} (acceptor)"
        )
        # Act depending on the number of bonds
        if len(donor.bonds) == 1:  # No H or LP attached
            self.make_atom_with_one_bond_h(donor, newname)
            newatom = donor.residue.get_atom(newname)
            return self.try_single_alcoholic_h(donor, acc, newatom)
        elif len(donor.bonds) == 2:
            loc1, loc2 = self.get_positions_with_two_bonds(donor)
            return self.try_positions_with_two_bonds_h(
                donor, acc, newname, loc1, loc2
            )
        elif len(donor.bonds) == 3:
            loc = self.get_position_with_three_bonds(donor)
            return self.try_positions_three_bonds_h(donor, acc, newname, loc)
        return False

    def try_acceptor(self, acc, donor):
        """Add a lone pair to an optimizeable residue.

        :param donor:  hydrogen bond donor
        :type donor:  Atom
        :param acc:  hydrogen bond acceptor
        :type acc:  Atom
        :return:  indication of whether addition was successful
        :rtype:  bool
        """
        residue = acc.residue
        # Do some error checking
        if not donor.hdonor:
            return False
        # Get the name of the LP to add
        if residue.has_atom("LP2"):
            return False
        elif residue.has_atom("LP1"):
            newname = "LP2"
        else:
            newname = "LP1"
        _LOGGER.debug(
            f"Working on {acc.residue} {acc.name} (acceptor) "
            f"to {donor.residue} {donor.name} (donor)"
        )
        # Act depending on the number of bonds
        if len(acc.bonds) == 1:  # No H or LP attached
            self.make_atom_with_one_bond_lp(acc, newname)
            newatom = acc.residue.get_atom(newname)
            return self.try_single_alcoholic_lp(acc, donor, newatom)
        elif len(acc.bonds) == 2:
            loc1, loc2 = self.get_positions_with_two_bonds(acc)
            return self.try_positions_with_two_bonds_lp(
                acc, donor, newname, loc1, loc2
            )
        elif len(acc.bonds) == 3:
            loc = self.get_position_with_three_bonds(acc)
            return self.try_positions_three_bonds_lp(acc, donor, newname, loc)
        return False

    def finalize(self):
        """Finalize an alcoholic residue.

        .. todo::
           Replace hard-coded values in the function.

        Try to minimize conflict with nearby atoms by building away from them.
        Called when LPs are still present so as to account for their bonds.
        """
        # Initialize some variables
        residue = self.residue
        atom = self.atomlist[0]
        # Conditions for return
        addname = self.hname
        if residue.fixed:
            return
        if residue.has_atom(addname):
            return
        if len(atom.bonds) == 1:
            # Initialize variables
            pivot = atom.bonds[0]
            # bestdist = 0.0
            bestcoords = []
            bestenergy = 999.99
            # Add atom and debump
            self.make_atom_with_one_bond_h(atom, addname)
            newatom = residue.get_atom(addname)
            self.routines.cells.add_cell(newatom)
            for _ in range(18):
                residue.rotate_tetrahedral(pivot, atom, 20.0)
                closeatoms = self.routines.cells.get_near_cells(atom)
                energy = 0.0
                for catom in closeatoms:
                    energy += self.get_pair_energy(
                        atom, catom
                    ) + self.get_pair_energy(catom, atom)
                if energy < bestenergy:
                    bestenergy = energy
                    bestcoords = newatom.coords
            if bestcoords != []:
                newatom.x = bestcoords[0]
                newatom.y = bestcoords[1]
                newatom.z = bestcoords[2]
        elif len(atom.bonds) == 2:
            loc1, loc2 = self.get_positions_with_two_bonds(atom)
            residue.create_atom(addname, loc1)
            newatom = residue.get_atom(addname)
            self.routines.cells.add_cell(newatom)
            # Debump residue if necessary by trying the other location
            closeatoms = self.routines.cells.get_near_cells(atom)
            energy1 = 0.0
            for catom in closeatoms:
                energy1 += self.get_pair_energy(
                    atom, catom
                ) + self.get_pair_energy(catom, atom)
            # Place at other location
            self.routines.cells.remove_cell(newatom)
            newatom.x = loc2[0]
            newatom.y = loc2[1]
            newatom.z = loc2[2]
            self.routines.cells.add_cell(newatom)
            energy2 = 0.0
            for catom in closeatoms:
                energy2 += self.get_pair_energy(
                    atom, catom
                ) + self.get_pair_energy(catom, atom)
            # If this is worse, switch back
            if energy2 > energy1:
                self.routines.cells.remove_cell(newatom)
                newatom.x = loc1[0]
                newatom.y = loc1[1]
                newatom.z = loc1[2]
                self.routines.cells.add_cell(newatom)
        elif len(atom.bonds) == 3:
            loc = self.get_position_with_three_bonds(atom)
            residue.create_atom(addname, loc)
            self.routines.cells.add_cell(residue.get_atom(addname))

    def complete(self):
        """Complete an alcoholic optimization.

        Call :func:`finalize` and then remove all extra LP atoms.
        """
        # Initialize some variables
        residue = self.residue
        self.finalize()
        residue.fixed = 1
        # Remove all LP atoms
        atomlist = list(residue.atoms)
        for atom in atomlist:
            if atom.name.startswith("LP"):
                residue.remove_atom(atom.name)


class Water(optimize.Optimize):
    """The class for water residues."""

    def __init__(self, residue, optinstance, routines):
        """Initialize the water optimization class.

        :param residue:  the residue to flip
        :type residue:  Residue
        :param optinstance:  the optimization instance containing information
            about what to optimize
        :type optinstance:  OptimizationHandler
        :param routines:  debumping routines object
        :type routines:  Debump
        """
        self.optinstance = optinstance
        self.residue = residue
        self.routines = routines
        self.hbonds = []
        oxatom = residue.get_atom("O")
        if oxatom is None:
            raise KeyError(f"Unable to find oxygen atom in {residue}!")
        oxatom.hdonor = 1
        oxatom.hacceptor = 1
        self.atomlist = [oxatom]

    def try_both(self, donor, acc, accobj):
        """Attempt to optimize both the donor and the acceptor.

        If one is fixed, we only need to try one side.  Otherwise first try to
        satisfy the donor - if that's successful, try to satisfy the acceptor.
        An undo may be necessary if the donor is satisfied and the acceptor
        isn't.

        :param donor:  hydrogen bond donor
        :type donor:  Atom
        :param acc:  hydrogen bond acceptor
        :type acc:  Atom
        :param accobj:  a Flip-like hydrogen bond structure object
        :type accobj:  Flip
        :return:  indication of whether hydrogen was added
        :rtype:  bool
        """
        # If one residue if fixed, use other functions
        residue = donor.residue
        if donor.residue.fixed:
            return bool(accobj.try_acceptor(acc, donor))
        if acc.residue.fixed:
            return bool(self.try_donor(donor, acc))
        if self.try_donor(donor, acc):
            if accobj.try_acceptor(acc, donor):
                _LOGGER.debug("NET BOND SUCCESSFUL!")
                return True
            # We need to undo what we did to the donor
            _LOGGER.debug("REMOVED NET HBOND")
            if residue.has_atom("H2"):
                residue.remove_atom("H2")
            elif residue.has_atom("H1"):
                residue.remove_atom("H1")
            return False
        return False

    def try_acceptor(self, acc, donor):
        """The main driver for adding an LP to an optimizeable residue.

        .. todo::
           This looks like a bug: donorh ends up being the last item in
           donor.bonds.
           This may be fixed by setting a best_donorh to go with bestdist and
           using best_donorh in the function below

        :param donor:  hydrogen bond donor
        :type donor:  Atom
        :param acc:  hydrogen bond acceptor
        :type acc:  Atom
        :return:  indication of whether addition was successful
        :rtype:  bool
        """
        residue = acc.residue
        # Do some error checking
        if not donor.hdonor:
            return False
        # Get the name of the LP to add
        if residue.has_atom("LP2"):
            return False
        elif residue.has_atom("LP1"):
            newname = "LP2"
        else:
            newname = "LP1"
        _LOGGER.debug(
            f"Working on {acc.residue} {acc.name} (acceptor) "
            f"to {donor.residue} {donor.name} (donor)"
        )
        # Act depending on the number of bonds
        if len(acc.bonds) == 0:
            if self.is_hbond(donor, acc):
                # Find the best donor hydrogen and use that
                best_donorh = donor.bonds[0]
                best_dist = util.distance(acc.coords, best_donorh.coords)
                for donorh in donor.bonds[1:]:
                    dist = util.distance(acc.coords, donorh.coords)
                    if dist < best_dist:
                        best_dist = dist
                        best_donorh = donorh
                self.make_atom_with_no_bonds(acc, best_donorh, newname)
                _LOGGER.debug(f"Added {newname} to {acc.residue}")
                return True
            return False
        elif len(acc.bonds) == 1:  # No H or LP attached
            _LOGGER.debug(
                f"Trying to add {newname} to {acc.residue} with one bond"
            )
            self.make_water_with_one_bond(acc, newname)
            newatom = acc.residue.get_atom(newname)
            return self.try_single_alcoholic_lp(acc, donor, newatom)
        elif len(acc.bonds) == 2:
            _LOGGER.debug(
                f"Trying to add {newname} to {acc.residue} with two bonds"
            )
            loc1, loc2 = self.get_positions_with_two_bonds(acc)
            return self.try_positions_with_two_bonds_lp(
                acc, donor, newname, loc1, loc2
            )
        elif len(acc.bonds) == 3:
            _LOGGER.debug(
                f"Trying to add {newname} to {acc.residue} with three bonds"
            )
            loc = self.get_position_with_three_bonds(acc)
            return self.try_positions_three_bonds_lp(acc, donor, newname, loc)
        return False

    def try_donor(self, donor, acc):
        """Add hydrogen to an optimizable residue.

        :param donor:  hydrogen bond donor
        :type donor:  Atom
        :param acc:  hydrogen bond acceptor
        :type acc:  Atom
        :return:  indication of whether addition was successful
        :rtype:  bool
        """
        residue = self.residue
        # Do some error checking
        if not acc.hacceptor:
            return False
        # Get the name of the atom to add
        if residue.has_atom("H2"):
            return False
        elif residue.has_atom("H1"):
            newname = "H2"
        else:
            newname = "H1"
        _LOGGER.debug(
            f"Working on {donor.residue} {donor.name} (donor) "
            f"to {acc.residue} {acc.name} (acceptor)"
        )
        # Act depending on the number of bonds
        if len(donor.bonds) == 0:
            self.make_atom_with_no_bonds(donor, acc, newname)
            if self.is_hbond(donor, acc):
                return True
            self.routines.cells.remove_cell(residue.get_atom(newname))
            residue.remove_atom(newname)
            return False
        if len(donor.bonds) == 1:
            self.make_water_with_one_bond(donor, newname)
            newatom = donor.residue.get_atom(newname)
            return self.try_single_alcoholic_h(donor, acc, newatom)
        elif len(donor.bonds) == 2:
            loc1, loc2 = self.get_positions_with_two_bonds(donor)
            return self.try_positions_with_two_bonds_h(
                donor, acc, newname, loc1, loc2
            )
        elif len(donor.bonds) == 3:
            loc = self.get_position_with_three_bonds(donor)
            return self.try_positions_three_bonds_h(donor, acc, newname, loc)
        return False

    def finalize(self):
        """Finalize a water residue.

        Try to minimize conflict with nearby atoms by building away from them.
        Called when LPs are still present so as to account for their bonds.
        """
        residue = self.residue
        # Conditions for return
        if residue.fixed:
            _LOGGER.debug(f"Residue {residue} already fixed")
            return
        if residue.has_atom("H2"):
            _LOGGER.debug(f"Residue {residue} already has H2")
            return
        atom = residue.get_atom("O")
        if not residue.has_atom("H1"):
            addname = "H1"
        else:
            addname = "H2"
        _LOGGER.debug(
            f"Finalizing {residue} by adding {addname} ({len(atom.bonds)} "
            "current O bonds)"
        )
        if len(atom.bonds) == 0:
            newcoords = []
            # Build hydrogen away from closest atom
            closeatom = self.routines.get_closest_atom(atom)
            if closeatom is not None:
                vec = util.subtract(atom.coords, closeatom.coords)
                dist = util.distance(atom.coords, closeatom.coords)
                for i in range(3):
                    newcoords.append(vec[i] / dist + atom.coords[i])
            else:
                newcoords = util.add(atom.coords, [1.0, 0.0, 0.0])
            residue.create_atom(addname, newcoords)
            self.routines.cells.add_cell(residue.get_atom(addname))
            self.finalize()
        elif len(atom.bonds) == 1:
            # Initialize variables
            pivot = atom.bonds[0]
            bestdist = 0.0
            bestcoords = []
            # Add atom and debump
            self.make_water_with_one_bond(atom, addname)
            newatom = residue.get_atom(addname)
            self.routines.cells.add_cell(newatom)
            for i in range(18):
                residue.rotate_tetrahedral(pivot, atom, 20.0)
                nearatom = self.routines.get_closest_atom(newatom)
                # If there is no closest conflict, continue
                if nearatom is None:
                    continue
                dist = util.distance(nearatom.coords, newatom.coords)
                if dist > bestdist:
                    bestdist = dist
                    bestcoords = newatom.coords
            if bestcoords != []:
                newatom.x = bestcoords[0]
                newatom.y = bestcoords[1]
                newatom.z = bestcoords[2]
            if addname == "H1":
                self.finalize()
            residue.fixed = 1
        elif len(atom.bonds) == 2:
            loc1, loc2 = self.get_positions_with_two_bonds(atom)
            residue.create_atom(addname, loc1)
            newatom = residue.get_atom(addname)
            self.routines.cells.add_cell(newatom)
            # Debump residue if necessary by trying the other location
            nearatom = self.routines.get_closest_atom(newatom)
            if nearatom is not None:
                dist1 = util.distance(newatom.coords, nearatom.coords)
                # Place at other location
                self.routines.cells.remove_cell(atom)
                newatom.x = loc2[0]
                newatom.y = loc2[1]
                newatom.z = loc2[2]
                self.routines.cells.add_cell(atom)
                nearatom = self.routines.get_closest_atom(newatom)
                if nearatom is not None:
                    # If this is worse, switch back
                    if util.distance(newatom.coords, nearatom.coords) < dist1:
                        self.routines.cells.remove_cell(atom)
                        newatom.x = loc1[0]
                        newatom.y = loc1[1]
                        newatom.z = loc1[2]
                        self.routines.cells.add_cell(atom)
            if addname == "H1":
                self.finalize()
        elif len(atom.bonds) == 3:
            loc = self.get_position_with_three_bonds(atom)
            residue.create_atom(addname, loc)
            self.routines.cells.add_cell(residue.get_atom(addname))

    def complete(self):
        """Complete the water optimization process."""
        self.finalize()
        residue = self.residue

        atomlist = list(residue.atoms)
        for atom in atomlist:
            if atom.name.startswith("LP"):
                residue.remove_atom(atom.name)


class Carboxylic(optimize.Optimize):
    """The class for carboxylic residues"""

    def __init__(self, residue, optinstance, routines):
        """Initialize carboxylic optimization class.

        Initialize a case where the lone hydrogen atom can have four different
        orientations. Works similar to :func:`initializeFlip` by pre-adding
        the necessary atoms.

        This also takes into account that the carboxyl group has different
        bond lengths for the two C-O bonds - this is probably due to one bond
        being assigned as a C=O.  As a result hydrogens are only added to the
        C-O (longer) bond.

        :param residue:  the residue to flip
        :type residue:  Residue
        :param optinstance:  the optimization instance containing information
            about what to optimize
        :type optinstance:  OptimizationHandler
        :param routines:  debumping routines object
        :type routines:  Debump
        """
        # Initialize some variables
        self.optinstance = optinstance
        self.residue = residue
        self.routines = routines
        self.atomlist = []
        self.hbonds = []
        self.hlist = []
        hname2 = ""
        hname1 = ""
        for name in optinstance.map.keys():
            if name.endswith("2"):
                hname2 = name
            else:
                hname1 = name
        bondatom1 = residue.get_atom(optinstance.map[hname1].bond)
        bondatom2 = residue.get_atom(optinstance.map[hname2].bond)
        longflag = 0
        # If one bond in the group is significantly (0.05 A)
        # longer than the other, use that group only
        for pivotatom in bondatom1.bonds:
            if not pivotatom.is_hydrogen:
                the_pivatom = pivotatom
                break
        # WARNING: What if the_pivatom is not set?
        dist1 = util.distance(the_pivatom.coords, bondatom1.coords)
        dist2 = util.distance(the_pivatom.coords, bondatom2.coords)
        order = [hname1, hname2]
        if dist2 > dist1 and abs(dist1 - dist2) > 0.05:
            longflag = 1
            order = [hname2, hname1]
        elif dist1 > dist2 and abs(dist1 - dist2) > 0.05:
            longflag = 1
            order = [hname1, hname2]
        for hname in order:
            bondatom = residue.get_atom(optinstance.map[hname].bond)
            # First mirror the hydrogen about the same donor
            for dihedral in residue.reference.dihedrals:
                if dihedral.endswith(hname):
                    the_dihedral = dihedral
                    break
            # WARNING: What if the_dihedral is not set?
            anglenum = residue.reference.dihedrals.index(the_dihedral)
            if anglenum == -1:
                raise IndexError("Unable to find Carboxylic dihedral angle!")
            if residue.dihedrals[anglenum] is None:
                self.atomlist.append(bondatom)
                continue
            newangle = 180.0 + residue.dihedrals[anglenum]
            self.routines.set_dihedral_angle(residue, anglenum, newangle)
            hatom = residue.get_atom(hname)
            newcoords = hatom.coords
            # Flip back to return original atom
            newangle = 180.0 + residue.dihedrals[anglenum]
            self.routines.set_dihedral_angle(residue, anglenum, newangle)
            # Rename the original atom and rebuild the new atom
            residue.rename_atom(hname, f"{hname}1")
            newname = f"{hname}2"
            residue.create_atom(newname, newcoords)
            newatom = residue.get_atom(newname)
            self.routines.cells.add_cell(newatom)
            newatom.refdistance = hatom.refdistance
            # Set the bonds for the new atom
            if bondatom not in newatom.bonds:
                newatom.bonds.append(bondatom)
            if newatom not in bondatom.bonds:
                bondatom.bonds.append(newatom)
            # Break if this is the only atom to add
            self.atomlist.append(bondatom)
            self.hlist.append(residue.get_atom(f"{hname}1"))
            self.hlist.append(residue.get_atom(f"{hname}2"))
            if longflag:
                break
        residue.set_donors_acceptors()

    def try_both(self, donor, acc, accobj):
        """Attempt to optimize donor and acceptor.

        Called when both the donor and acceptor are optimizeable.
        If one is fixed, we only need to try one side.
        Otherwise first try to satisfy the donor - if that's succesful, try to
        satisfy the acceptor.
        An undo may be necessary if the donor is satisfied and the acceptor
        isn't.

        :param donor:  hydrogen bond donor
        :type donor:  Atom
        :param acc:  hydrogen bond acceptor
        :type acc:  Atom
        :param accobj:  a Flip-like hydrogen bond structure object
        :type accobj:  Flip
        :return:  indication of whether hydrogen was added
        :rtype:  bool
        """
        # If one residue if fixed, use other functions
        if donor.residue.fixed:
            return bool(accobj.try_acceptor(acc, donor))
        if acc.residue.fixed:
            return bool(self.try_donor(donor, acc))
        _LOGGER.debug(
            f"Working on {donor.residue} {donor.name} (donor) "
            f"to {acc.residue} {acc.name} (acceptor)"
        )
        if self.is_hbond(donor, acc):
            if accobj.try_acceptor(acc, donor):
                self.fix(donor, acc)
                _LOGGER.debug("NET BOND SUCCESSFUL!")
                return True
            return False
        return False

    def is_carboxylic_hbond(self, donor, acc):
        """Determine whether this donor acceptor pair is a hydrogen bond.

        :param donor:  hydrogen bond donor
        :type donor:  Atom
        :param acc:  hydrogen bond acceptor
        :type acc:  Atom
        :param accobj:  a Flip-like hydrogen bond structure object
        :type accobj:  Flip
        :return:  indication of whether hydrogen was added
        :rtype:  bool
        """
        for donorhatom in donor.bonds:
            if not donorhatom.is_hydrogen:
                continue
            # Check the H(D)-A distance
            dist = util.distance(donorhatom.coords, acc.coords)
            if dist > DIST_CUTOFF:
                continue
            # Check the A-D-H(D) angle
            angle = self.get_hbond_angle(acc, donor, donorhatom)
            if angle <= ANGLE_CUTOFF:
                _LOGGER.debug(f"Found HBOND! {dist:.4f} {angle:.4f}")
                return True
        # If we get here, no bond is formed
        return False

    def try_acceptor(self, acc, donor):
        """Add lone pair to an optimizable residue.

        :param donor:  hydrogen bond donor
        :type donor:  Atom
        :param acc:  hydrogen bond acceptor
        :type acc:  Atom
        :return:  indication of whether addition was successful
        :rtype:  bool
        """
        residue = acc.residue
        # Do some error checking
        if not donor.hdonor:
            return False
        _LOGGER.debug(
            f"Working on {acc.residue} {acc.name} (acceptor) "
            f"to {donor.residue} {donor.name} (donor)"
        )
        # We want to ignore the Hs on the acceptor
        if self.is_carboxylic_hbond(donor, acc):
            dist = None
            donorhatom = None
            # Eliminate the closer hydrogen
            hyds = [hatom for hatom in self.hlist if hatom.is_hydrogen]
            if len(hyds) < 2:
                return True
            dist = util.distance(hyds[0].coords, donor.coords)
            dist2 = util.distance(hyds[1].coords, donor.coords)
            # Eliminate hyds[0]
            if dist < dist2:
                self.hlist.remove(hyds[0])
                self.routines.cells.remove_cell(hyds[0])
                residue.remove_atom(hyds[0].name)
                donorhatom = residue.get_atom(hyds[1].name)
            elif hyds[1] in self.hlist:
                self.hlist.remove(hyds[1])
                self.routines.cells.remove_cell(hyds[1])
                residue.remove_atom(hyds[1].name)
                if residue.has_atom(hyds[0].name):
                    donorhatom = residue.get_atom(hyds[0].name)
                elif len(self.hlist) != 0 and residue.has_atom(
                    self.hlist[0].name
                ):
                    donorhatom = residue.get_atom(self.hlist[0].name)
            # If only one H is left, we're done
            if len(self.hlist) == 1:
                if donorhatom is not None:
                    self.rename(donorhatom)
                residue.fixed = 1
            return True
        else:
            return False

    def try_donor(self, donor, acc):
        """Add hydrogen to an optimizable residue.

        :param donor:  hydrogen bond donor
        :type donor:  Atom
        :param acc:  hydrogen bond acceptor
        :type acc:  Atom
        :return:  indication of whether addition was successful
        :rtype:  bool
        """
        # Do some error checking
        if not acc.hacceptor:
            return False
        if self.is_hbond(donor, acc):
            self.fix(donor, acc)
            return True
        return False

    def fix(self, donor, acc):
        """Fix the carboxylic residue."""
        _LOGGER.debug(f"Fixing residue {donor.residue} due to {donor.name}")
        residue = donor.residue
        # Grab the H(D) that caused the bond
        for donorhatom in donor.bonds:
            if donorhatom.is_hydrogen and (
                self.get_hbond_angle(acc, donor, donorhatom) <= ANGLE_CUTOFF
            ):
                the_donorhatom = donorhatom
                break
        # Remove all the other available bonded hydrogens
        hydrogens = self.hlist[:]
        for atom in hydrogens:
            if atom != the_donorhatom:
                self.routines.cells.remove_cell(atom)
                self.hlist.remove(atom)
                residue.remove_atom(atom.name)
        # Rename the atoms
        # WARNING: Nathan? What if the_donorhatom is not set?
        self.rename(the_donorhatom)
        residue.fixed = 1

    def finalize(self):
        """Finalize a protontated residue.

        Try to minimize conflict with nearby atoms.
        """
        # Initialize some variables
        hydrogens = []
        bestatom = None
        residue = self.residue
        if residue.fixed:
            return
        # For each atom, get the closest atom
        bestenergy = 999.99
        for hydatom in self.hlist:
            energy = 0
            bondedatom = hydatom.bonds[0]
            closeatoms = self.routines.cells.get_near_cells(bondedatom)
            for catom in closeatoms:
                energy += self.get_pair_energy(
                    bondedatom, catom
                ) + self.get_pair_energy(catom, bondedatom)
            if energy < bestenergy:
                bestenergy = energy
                bestatom = hydatom
        # Keep the bestatom
        for hydatom in self.hlist:
            hydrogens.append(hydatom)
        for hydatom in hydrogens:
            if bestatom != hydatom:
                self.hlist.remove(hydatom)
                residue.remove_atom(hydatom.name)
        # Rename the atoms
        if bestatom is not None and len(bestatom.name) == 4:
            self.rename(bestatom)
        else:
            pass
        residue.fixed = 1

    def rename(self, hydatom):
        """Rename the optimized atoms appropriately.

        This is done since the forcefields tend to require that the hydrogen
        is linked to a specific oxygen, and this atom may have different
        parameter values.

        :param hydatom:  the hydrogen atom that was added
        :type hydatom:  Atom
        """
        residue = self.residue
        optinstance = self.optinstance
        # No need to rename if hydatom is not in residue.map
        if hydatom.name not in residue.map.keys():
            return
        # Take off the extension
        if len(hydatom.name) == 4:
            hname = hydatom.name[:-1]
            residue.rename_atom(hydatom.name, hname)
        # PATCHES.xml expects *2 - if it's *1 that left, flip names
        if len(self.atomlist) == 2:
            if hydatom.name.endswith("1"):
                residue.rename_atom(hydatom.name, f"{hydatom.name[:-1]}2")
                bondname0 = self.atomlist[0].name
                bondname1 = self.atomlist[1].name
                tempname = "FLIP"
                residue.rename_atom(self.atomlist[0].name, tempname)
                residue.rename_atom(self.atomlist[1].name, bondname0)
                residue.rename_atom(tempname, bondname1)
            elif hydatom.name.endswith("OXT"):
                residue.rename_atom(hydatom.name, "HO")
                bondname0 = self.atomlist[0].name
                bondname1 = self.atomlist[1].name
                tempname = "FLIP"
                residue.rename_atom(self.atomlist[0].name, tempname)
                residue.rename_atom(self.atomlist[1].name, bondname0)
                residue.rename_atom(tempname, bondname1)
        elif len(self.atomlist) == 1:
            # Appending the other bondatom to self.atomlist
            # WARNING: Nathan? What if hname is not set?
            hnames = [hname[:-1] + "1", hname[:-1] + "2"]
            for hname in hnames:
                bondatom = residue.get_atom(optinstance.map[hname].bond)
                if bondatom.name != self.atomlist[0].name:
                    self.atomlist.append(bondatom)
            if hydatom.name.endswith("1"):
                if hydatom.name[:-1] + "2" in residue.map.keys():
                    residue.remove_atom(f"{hydatom.name[:-1]}2")
                residue.rename_atom(hydatom.name, f"{hydatom.name[:-1]}2")
                bondname0 = self.atomlist[0].name
                bondname1 = self.atomlist[1].name
                tempname = "FLIP"
                residue.rename_atom(self.atomlist[0].name, tempname)
                residue.rename_atom(self.atomlist[1].name, bondname0)
                residue.rename_atom(tempname, bondname1)
            elif hydatom.name.endswith("OXT"):
                residue.rename_atom(hydatom.name, "HO")
                bondname0 = self.atomlist[0].name
                bondname1 = self.atomlist[1].name
                tempname = "FLIP"
                residue.rename_atom(self.atomlist[0].name, tempname)
                residue.rename_atom(self.atomlist[1].name, bondname0)
                residue.rename_atom(tempname, bondname1)

    def complete(self):
        """If not already fixed, finalize the optimization."""
        if (len(self.hlist) == 2) and (self.residue.fixed == 1):
            self.residue.fixed = 0
        if not self.residue.fixed:
            self.finalize()


class PotentialBond:
    """A class containing the hydrogen bond structure."""

    def __init__(self, atom1, atom2, dist):
        """Initialize the class.

        :param atom1:  the first atom in the potential bond
        :type atom1:  Atom
        :param atom2:  the second atom in the potential bond
        :type atom2:  Atom
        :param dist:  the distance between the two atoms
        :type dist:  float
        """
        self.atom1 = atom1
        self.atom2 = atom2
        self.dist = dist

    def __str__(self):
        txt = f"{self.atom1.name} {self.atom1.residue}"
        txt += " to "
        txt += f"{self.atom2.name} {self.atom2.residue}"
        txt += f" ({self.dist:.2f} A)"
        return txt


class HydrogenDefinition:
    """Class for potential ambiguities in amino acid hydrogens.

    It is essentially the hydrogen definition file in object form.
    """

    def __init__(self, name, opttype, optangle, map_):
        """Initialize the object with information from the definition file.

        See :file:`HYDROGENS.XML` for more information.

        :param name:  the name of the grouping
        :type name:  str
        :param opttype:  the optimization type of the grouping
        :type opttype:  str
        :param optangle:  the optimization angle of the grouping
        :type optangle:  str
        :param map:  the map of hydrogens
        """
        self.name = name
        self.opttype = opttype
        self.optangle = optangle
        self.map = map_
        self.conformations = []

    def __str__(self):
        output = f"Name:                  {self.name}\n"
        output += f"Opttype:               {self.opttype}\n"
        output += f"Optangle:              {self.optangle}\n"
        output += f"map:                   {self.map}\n"
        output += "Conformations:\n"
        for conf in self.conformations:
            output += f"\n{conf}"
        output += "*****************************************\n"
        return output

    def add_conf(self, conf):
        """Add a hydrogen conformation to the list.

        :param conf:  the conformation to be added
        :type conf:  HydrogenConformation
        """
        self.conformations.append(conf)


class HydrogenConformation:
    """Class for possible hydrogen conformations.

    The :class:`HydrogenConformation` class contains data about possible
    hydrogen conformations as specified in the hydrogen data file.
    """

    def __init__(self, hname: str, boundatom: str, bondlength: float):
        """
        Args:
            hname: The hydrogen name
            boundatom: The atom the hydrogen is bound to
            bondlength: The bond length
        """
        self.hname = hname
        self.boundatom = boundatom
        self.bondlength = bondlength
        self.atoms = []

    def __str__(self):
        output = f"Hydrogen Name: {self.hname}\n"
        output += f"Bound Atom:    {self.boundatom}\n"
        output += f"Bond Length:   {self.bondlength:.2f}\n"
        for atom in self.atoms:
            output += f"\t{atom}\n"
        return output

    def add_atom(self, atom):
        """Add an atom to the list of atoms.

        :param atom:  the atom to be added
        :type atom:  DefinitionAtom
        """
        self.atoms.append(atom)


class HydrogenAmbiguity:
    """Contains information about ambiguities in hydrogen conformations."""

    def __init__(self, residue, hdef, routines):
        """If the residue has a rotateable hydrogen, remove it.

        If it can be flipped, pre-flip the residue by creating all additional
        atoms.

        :param residue:  the residue in question
        :type residue:  Residue
        :param hdef:  the hydrogen definition matching the residue
        :param routines:  the debumping routines object
        :type routines:  Debump
        """
        self.residue = residue
        self.hdef = hdef
        self.routines = routines

    def __str__(self):
        text = f"{self.residue.name} {self.residue.res_seq} "
        text += f"{self.residue.chain_id} ({self.hdef.opttype})"
        return text


class HydrogenHandler(sax.ContentHandler):
    """Extends the SAX XML Parser to parse the Hydrogens.xml class."""

    def __init__(self):
        self.curelement = ""
        self.curatom = None
        self.curobj = None
        self.curholder = None
        self.map = {}

    def startElement(self, name, _):
        """Create optimization holder objects or atoms.

        .. todo::
           Rename this and related methods to conform with PEP8

        :param name:  name for element
        :type name:  str
        """
        if name == "class":
            obj = optimize.OptimizationHolder()
            self.curholder = obj
            self.curobj = obj
        elif name == "atom":
            obj = defns.DefinitionAtom()
            self.curatom = obj
            self.curobj = obj
        else:
            self.curelement = name

    def endElement(self, name):
        """Complete object is passed in by name parameter.

        .. todo::
           Rename this and related methods to conform with PEP8

        :param name:  name for element
        :type name:  str
        """
        if name == "class":  # Complete Residue object
            obj = self.curholder
            if not isinstance(obj, optimize.OptimizationHolder):
                raise ValueError("Internal error parsing XML!")
            self.map[obj.name] = obj
            self.curholder = None
            self.curobj = None
        elif name == "atom":  # Complete atom object
            atom = self.curatom
            if not isinstance(atom, defns.DefinitionAtom):
                raise ValueError("Internal error parsing XML!")
            atomname = atom.name
            if atomname == "":
                raise ValueError("Atom name not set in XML!")
            else:
                self.curholder.map[atomname] = atom
                self.curatom = None
                self.curobj = self.curholder
        else:  # Just free the current element namespace
            self.curelement = ""
        return self.map

    def characters(self, text):
        """Set a given attribute of the object to the text.

        :param text:  value of the attribute
        :type text:  str
        """
        if text.isspace():
            return
        # If this is a float, make it so
        try:
            value = float(str(text))
        except ValueError:
            value = str(text)
        setattr(self.curobj, self.curelement, value)



"""Utilities for the PDB2PQR software suite.

.. todo:
    The functions in this module are great examples of why PDB2PQR needs
    :mod:`numpy`.
    More efforts should be made to subsitute with :mod:`numpy` data types and
    functions wherever possible throughout the code base.

.. codeauthor::  Todd Dolinsky
.. codeauthor::  Yong Huang
.. codeauthor::  Nathan Baker
"""

import logging
import math

import numpy as np

from .config import CHARGE_ERROR, RADIANS_TO_DEGREES, SMALL_NUMBER

_LOGGER = logging.getLogger(__name__)


def noninteger_charge(charge, error_tol=CHARGE_ERROR) -> str:
    """Test whether a charge is an integer.

    :param float charge:  value to test
    :param float error_tol:  absolute error tolerance
    :returns:  string with descripton of problem or empty string if no problem
    """
    abs_error = abs(charge - round(charge))
    if abs_error > abs(error_tol):
        return (
            f"{charge} deviates by {abs_error} from integral, exceeding error "
            f"tolerance {error_tol}"
        )
    return ""


def sort_dict_by_value(inputdict):
    """Sort a dictionary by its values.

    :param inputdict:  the dictionary to sort
    :type inputdict:  dict
    :return:  list of keys sorted by value
    :rtype:  list
    """
    items = sorted(inputdict.items(), key=lambda x: x[1], reverse=True)
    items = [v for v, k in items]
    return items


def shortest_path(graph, start, end, path=[]):
    """Find the shortest path between two nodes.

    Uses recursion to find the shortest path from one node to another in an
    unweighted graph.
    Adapted from http://www.python.org/doc/essays/graphs.html

    :param graph:  a mapping of the graph to analyze, of the form {0: [1,2],
        1:[3,4], ...} . Each key has a list of edges.
    :type graph:  dict
    :param start:  the ID of the key to start the analysis from
    :type start:  str
    :param end:  the ID of the key to end the analysis
    :type end:  str
    :param path:  optional argument used during the recursive step to keep
        the current path up to that point
    :type path:  list

    :return:  list of the shortest path or ``None`` if start and end are not
        connected
    :rtype:  list
    """
    path = [*path, start]
    if start == end:
        return path
    if start not in graph:
        return None
    shortest = None
    for node in graph[start]:
        if node not in path:
            newpath = shortest_path(graph, node, end, path)
            if newpath and (not shortest or len(newpath) < len(shortest)):
                shortest = newpath
    return shortest


def analyze_connectivity(map_, key):
    """Analyze the connectivity of a given map using the key value.

    :param map:  map to analyze
    :type map:  dict
    :param key:  key value
    :type key:  str
    :return:  list of connected values to the key
    :rtype:  list
    """
    clist = []
    keys = [key]
    while keys:
        key = keys[0]
        if key not in clist:
            clist.append(key)
            if key in map_:
                for value in map_[key]:
                    if value not in clist:
                        keys.append(value)
        keys.pop(keys.index(key))
    return clist


def angle(coords1, coords2, coords3):
    """Get the angle between three coordinates.

    :param coords1:  first coordinate set
    :type coords1:  [float, float, float]
    :param coords2:  second (vertex) coordinate set
    :type coords2:  [float, float, float]
    :param coords3:  third coordinate set
    :type coords3:  [float, float, float]
    :return:  angle between the atoms (in degrees)
    :rtype:  float
    """
    diff32 = np.array(coords3) - np.array(coords2)
    diff12 = np.array(coords1) - np.array(coords2)
    norm1 = normalize(diff32)
    norm2 = normalize(diff12)
    dotted = np.inner(norm1, norm2)
    if dotted > 1.0:  # If normalized, this is due to rounding error
        dotted = 1.0
    elif dotted < -1.0:
        dotted = -1.0
    rad = np.absolute(np.arccos(dotted))
    value = rad * 180.0 / np.pi
    if value > 180.0:
        value = 360.0 - value
    return value


def distance(coords1, coords2):
    """Calculate the distance between two coordinates.

    :param coords1:  coordinates of form [x,y,z]
    :type coords1:  [float, float, float]
    :param coords2:  coordinates of form [x,y,z]
    :type coords2:  [float, float, float]

    :return:  distance between the two coordinates
    :rtype:  float
    """
    coords1 = np.array(coords1)
    coords2 = np.array(coords2)
    return np.linalg.norm(coords1 - coords2)


def add(coords1, coords2):
    """Add one 3-dimensional point to another.

    :param coords1:  coordinates of form [x,y,z]
    :type coords1:  [float, float, float]
    :param coords2:  coordinates of form [x,y,z]
    :type coords2:  [float, float, float]
    :return:  list of coordinates equal to coords2 + coords1
    :rtype:  numpy.ndarray
    """
    return np.array(coords1) + np.array(coords2)


def subtract(coords1, coords2):
    """Suntract one 3-dimensional point from another.

    :param coords1:  coordinates of form [x,y,z]
    :type coords1:  [float, float, float]
    :param coords2:  coordinates of form [x,y,z]
    :type coords2:  [float, float, float]
    :return:  list of coordinates equal to coords2 - coords1
    :rtype:  numpy.ndarray
    """
    return np.array(coords1) - np.array(coords2)


def cross(coords1, coords2):
    """Find the cross-product of one 3-dimensional point with another.

    :param coords1:  coordinates of form [x,y,z]
    :type coords1:  [float, float, float]
    :param coords2:  coordinates of form [x,y,z]
    :type coords2:  [float, float, float]
    :return:  list of coordinates equal to coords2 cross coords1
    :rtype:  numpy.ndarray
    """
    return np.cross(np.array(coords1), np.array(coords2))


def dot(coords1, coords2):
    """Find the dot-product of one 3-dimensional point with another.

    :param coords1:  coordinates of form [x,y,z]
    :type coords1:  [float, float, float]
    :param coords2:  coordinates of form [x,y,z]
    :type coords2:  [float, float, float]
    :return:  list of coordinates equal to the inner product of coords2 with
        coords1
    :rtype:  numpy.ndarray
    """
    return np.inner(np.array(coords1), np.array(coords2))


def normalize(coords):
    """Normalize a set of coordinates to unit vector.

    :param coords:  coordinates of form [x,y,z]
    :type coords:  [float, float, float]
    :return: normalized coordinates
    :rtype:  numpy.ndarray
    """
    return coords / np.linalg.norm(coords)


def factorial(num):
    """Returns the factorial of the given number.

    :param num:  number for which to compute factorial
    :type num:  int
    :return:  factorial of number
    :rtype:  int
    """
    if num <= 1:
        return 1
    return num * factorial(num - 1)


def dihedral(coords1, coords2, coords3, coords4):
    """Calculate the dihedral angle from four atoms' coordinates.

    :param coords1:  one of four coordinates of form [x,y,z]
    :type coords1:  [float, float, float]
    :param coords2:  one of four coordinates of form [x,y,z]
    :type coords2:  [float, float, float]
    :param coords3:  one of four coordinates of form [x,y,z]
    :type coords3:  [float, float, float]
    :param coords4:  one of four coordinates of form [x,y,z]
    :type coords4:  [float, float, float]
    :return:  the angle (in degrees)
    :rtype:  float
    """
    diff43 = np.array(coords4) - np.array(coords3)
    diff32 = np.array(coords3) - np.array(coords2)
    diff12 = np.array(coords1) - np.array(coords2)
    c12_32 = np.cross(diff12, diff32)
    c12_32 = normalize(c12_32)
    c43_32 = np.cross(diff43, diff32)
    c43_32 = normalize(c43_32)
    scal = np.inner(c12_32, c43_32)
    if np.absolute(scal + 1.0) < SMALL_NUMBER:
        value = 180.0
    elif np.absolute(scal - 1.0) < SMALL_NUMBER:
        value = 0.0
    else:
        value = RADIANS_TO_DEGREES * math.acos(scal)
    chiral = np.inner(np.cross(c12_32, c43_32), diff32)
    if chiral < 0:
        value *= -1.0
    return value



import re

import pdb2pqr
from pdb2pqr import __version__


def test_version_exists():
    assert hasattr(pdb2pqr, "__version__")


def test_version():
    assert re.match(r"[0-9]+\.[0-9]+\.[0-9]+", __version__)



"""Support molecules in Tripos MOL2 format.

For further information look at (web page exists: 25 August 2005):
http://www.tripos.com/index.php?family=modules,SimplePage,,,&page=sup_mol2&s=0
"""

import logging
from collections import OrderedDict
from itertools import combinations

from numpy import array
from numpy.linalg import norm

from . import NONBONDED_BY_TYPE, RADII, VALENCE_BY_ELEMENT, peoe

_LOGGER = logging.getLogger(__name__)


# These are the allowed bond types
BOND_TYPES = {"single", "double", "triple", "aromatic"}
# This is the maximum deviation from an ideal bond distance
BOND_DIST = 2.0


class Mol2Bond:
    """MOL2 molecule bonds."""

    def __init__(self, atom1, atom2, bond_type, bond_id=0):
        """Initialize bond.

        :param atom1:  name of first atom in bond
        :type atom1:  str
        :param atom2:  name of second atom in bond
        :type atom2:  str
        :param bond_type:  type of bond:  1 (single), 2 (double), or ar
            (aromatic)
        :type bond_type:  int
        :param bond_id:  integer ID of bond
        :type bond_id:  int
        """
        self.atoms = (atom1, atom2)
        self.bond_id = int(bond_id)
        if bond_type in BOND_TYPES:
            self.type = bond_type
        else:
            err = f"Unknown bond type: {bond_type}"
            raise ValueError(err)

    @property
    def atom_names(self):
        """Get atom names in bond.

        :return:  tuple with names of atoms in bond.
        :rtype:  (str, str)
        """
        return (self.atoms[0].name, self.atoms[1].name)

    @property
    def length(self):
        """Get bond length.

        :return:  bond length
        :rtype:  float
        """
        return self.atoms[0].distance(self.atoms[1])

    def __str__(self):
        mol2 = f"{self.atoms[0].name:s} {self.type:s}-bonded to "
        mol2 += f"{self.atoms[1].name:s}"
        return mol2


class Mol2Atom:
    """MOL2 molecule atoms."""

    def __init__(self):
        self.serial = None
        self.name = None
        self.alt_loc = None
        self.res_name = None
        self.chain_id = None
        self.res_seq = None
        self.x = None
        self.y = None
        self.z = None
        self.type = None
        self.radius = None
        self.is_c_term = False
        self.is_n_term = False
        self.mol2charge = None
        self.occupancy = 0.00
        self.temp_factor = 0.00
        self.seg_id = None
        self.charge = None
        self.num_rings = 0
        self.radius = None
        self.bonded_atoms = []
        self.bonds = []
        self.torsions = []
        self.rings = []
        # Terms for calculating atom electronegativity
        self.poly_terms = None
        # Atom electronegativity
        self.chi = None
        # Atom charge change during equilibration
        self.delta_charge = None

    def distance(self, other):
        """Get distance between two atoms.

        :param other:  other atom for distance measurement
        :type other:  Mol2Atom
        :return:  distance
        :rtype:  float
        """
        return norm(other.coords - self.coords)

    def __str__(self):
        """Generate PDB line from MOL2."""
        mol2 = (
            f"HETATM{self.serial:5d}{self.name:>5s}{self.res_name:>4s} L"
            f"{self.res_seq!s:>5s}   {self.x:8.3f}{self.y:8.3f}{self.z:8.3f}"
        )
        return mol2

    def assign_radius(self, primary_dict, secondary_dict):
        """Assign radius to atom.

        .. todo::
           It seems inconsistent that this function pulls radii from a
           dictionary and the biomolecule routines use force field files.

        :param primary_dict:  primary dictionary of radii indexed by atom
            type or element
        :type primary_dict:  dict
        :param secondary_dict:  backup dictionary for radii not found in
            primary dictionary
        :type secondary_dict:  dict
        """
        radius = None
        for rdict in [primary_dict, secondary_dict]:
            if radius is not None:
                break
            for key in [self.type, self.element]:
                if key in rdict:
                    radius = rdict[key]
                    break
        if radius is not None:
            self.radius = radius
        else:
            err = (
                f"Unable to find radius parameter for self of type "
                f"{self.type} in radius dictionary: {primary_dict}"
            )
            raise KeyError(err)

    @property
    def coords(self):
        """Coordinates.

        :return:  coordinates
        :rtype:  numpy.ndarray
        """
        return array([self.x, self.y, self.z])

    @property
    def bonded_atom_names(self):
        """Bonded atom names.

        :return:  bonded atom names
        :rtype:  list
        """
        return [a.name for a in self.bonded_atoms]

    @property
    def num_bonded_heavy(self):
        """Number of heavy atoms bonded to this atom.

        :return:  number of heavy atoms
        :rtype: int
        """
        return len([a for a in self.bonded_atoms if a.type != "H"])

    @property
    def num_bonded_hydrogen(self):
        """Number of hydrogen atoms bonded to this atom.

        :return:  number of hydrogen atoms
        :rtype:  int
        """
        return len([a for a in self.bonded_atoms if a.type == "H"])

    @property
    def element(self):
        """Element for this atom (uppercase).

        :return:  element for this atom
        :rtype:  str
        """
        return self.type.split(".")[0].upper()

    @property
    def bond_order(self):
        """Total number of electrons in bonds with other atoms.

        :return:  total number of electrons in bonds with other atoms
        :rtype:  int
        """
        order = 0
        num_aromatic = 0
        for bond in self.bonds:
            if bond.type == "single":
                order += 1
            elif bond.type == "double":
                order += 2
            elif bond.type == "triple":
                order += 3
            elif bond.type == "aromatic":
                num_aromatic += 1
            else:
                err = f"Unknown bond type: {bond.type}"
                raise ValueError(err)
        if num_aromatic > 0:
            order = order + num_aromatic + 1
        return order

    @property
    def formal_charge(self):
        """Formal charge for this atom

        :return:  formal charge for this atom
        :rtype:  int
        """
        element = self.type.split(".")[0]
        valence = VALENCE_BY_ELEMENT[element]
        nonbonded = NONBONDED_BY_TYPE[self.type]
        bond_order = self.bond_order
        formal_charge = valence - nonbonded - bond_order
        if (
            (self.type in ["N.pl3", "N.am"])
            and (bond_order == 3)
            and (formal_charge != 0)
        ):
            # Planar nitrogen bond orders are not always correct in MOL2
            _LOGGER.warning("Correcting planar/amide bond order.")
            formal_charge = 0
        elif (
            (self.type in ["N.ar"])
            and (bond_order == 4)
            and (formal_charge != 0)
        ):
            # Aromatic nitrogen bond orders are not always correct in MOL2
            _LOGGER.warning("Correcting aromatic nitrogen bond order.")
            formal_charge = 0
        elif (
            (self.type in ["C.ar"])
            and (bond_order == 5)
            and (formal_charge != 0)
        ):
            # Aromatic carbon bond orders are not always correct in MOL2
            _LOGGER.warning("Correcting aromatic carbon bond order.")
            formal_charge = 0
        elif (
            (self.type in ["O.co2"])
            and (bond_order == 1)
            and (formal_charge != -0.5)
        ):
            # CO2 bond orders are hardly ever set correctly in MOL2
            formal_charge = -0.5
        elif (
            (self.type in ["C.2"])
            and (bond_order == 5)
            and (formal_charge == -1)
        ):
            # CO2 bond orders are hardly ever set correctly in MOL2
            formal_charge = 0
        elif (
            (self.type in ["N.3"])
            and (bond_order == 4)
            and (formal_charge == -1)
        ):
            # Tetravalent nitrogen atom types are sometimes wrong in MOL2
            _LOGGER.warning("Correcting ammonium atom type.")
            formal_charge = 1
        elif (
            (self.type in ["O.3"])
            and (bond_order == 1)
            and (formal_charge == 1)
        ):
            # Phosphate groups are sometimes confused in MOL2
            # Assign negative charge to first O.3 with bond order 1
            # attached to phosphorous
            elements = [a.type[0] for a in self.bonds[0].atoms]
            p_atom = self.bonds[0].atoms[elements.index("P")]
            _LOGGER.warning("Correcting phosphate bond order.")
            o_atoms = []
            for bond in p_atom.bonds:
                for atom in bond.atoms:
                    if atom.type[0] == "O" and atom.bond_order == 1:
                        o_atoms.append(atom.name)
            formal_charge = -1 if o_atoms.index(self.name) == 0 else 0
        return formal_charge


class Mol2Molecule:
    """Tripos MOL2 molecule."""

    def __init__(self):
        self.atoms = OrderedDict()
        self.bonds = []
        self.torsions = set()
        self.rings = set()
        self.serial = None
        self.name = None
        self.res_name = None
        self.res_seq = None

    def assign_parameters(
        self, primary_dict=RADII["zap9"], secondary_dict=RADII["bondi"]
    ):
        """Assign charges and radii to atoms in molecule.

        Args:
            primary_dict:  primary dictionary of radii indexed by atom type or
                           element
            secondary_dict:  backup dictionary for radii not found in primary
                             dictionary
        """
        self.assign_radii(primary_dict, secondary_dict)
        self.assign_charges()

    def assign_radii(self, primary_dict, secondary_dict):
        """Assign radii to atoms in molecule.

        :param primary_dict:  primary dictionary of radii indexed by atom
            type or element
        :type primary_dict:  dict
        :param secondary_dict:  backup dictionary for radii not found in
            primary dictionary
        :type secondary_dict:  dict
        """
        for atom in self.atoms.values():
            atom.assign_radius(primary_dict, secondary_dict)

    def assign_charges(self):
        """Assign charges to atoms in molecule."""
        for atom in self.atoms.values():
            atom.charge = atom.formal_charge
        peoe.equilibrate(self.atoms.values())

    def find_atom_torsions(self, start_atom):
        """Set the torsion angles that start with this atom (name).

        :param start_atom:  starting atom name
        :type start_atom:  str
        :return: list of 4-tuples containing atom names comprising torsions
        """
        torsions = []
        for bonded1 in self.atoms[start_atom].bonded_atom_names:
            for bonded2 in self.atoms[bonded1].bonded_atom_names:
                if bonded2 == start_atom:
                    continue
                for end_atom in self.atoms[bonded2].bonded_atom_names:
                    if end_atom == bonded1:
                        continue
                    torsions.append((start_atom, bonded1, bonded2, end_atom))
        return torsions

    def set_torsions(self):
        """Set all torsions in molecule."""
        for atom_name, atom in self.atoms.items():
            atom.torsions = self.find_atom_torsions(atom_name)
            for torsion in atom.torsions:
                self.torsions.add(torsion)

    @staticmethod
    def rotate_to_smallest(path):
        """Rotate cycle path so that it begins with the smallest node.

        This was borrowed from StackOverflow: https://j.mp/2AHaukj

        :param path:  list of atom names
        :type path:  list of str
        :return:  rotated path
        :rtype:  list of str
        """
        n = path.index(min(path))
        return path[n:] + path[:n]

    def find_new_rings(self, path, rings, level=0):
        """Find new rings in molecule.

        This was borrowed from StackOverflow: https://j.mp/2AHaukj

        :param path:  list of atom names
        :type path:  list of str
        :param rings:  current list of rings
        :type rings:  list of str
        :param level:  recursion level
        :type level:  int
        :return:  new list of rings
        :rtype:  int
        """
        start_node = path[0]
        next_node = None
        sub_path = []
        for bond in self.bonds:
            atom1 = bond.atoms[0].name
            atom2 = bond.atoms[1].name
            if start_node in (atom1, atom2):
                next_node = atom2 if atom1 == start_node else atom1
                if next_node not in path:
                    sub_path = [next_node, *path]
                    rings = self.find_new_rings(sub_path, rings, level + 1)
                elif len(path) > 2 and next_node == path[-1]:
                    path_ = self.rotate_to_smallest(path)
                    inv_path = tuple(self.rotate_to_smallest(path_[::-1]))
                    path_ = tuple(path_)
                    if (path_ not in rings) and (inv_path not in rings):
                        rings.add(tuple(path_))
        return rings

    def set_rings(self):
        """Set all rings in molecule.

        This was borrowed from StackOverflow: https://j.mp/2AHaukj
        """
        self.rings = set()
        rings = set()
        # Generate all rings
        for bond in self.bonds:
            for atom_name in bond.atom_names:
                rings = self.find_new_rings([atom_name], rings)
        # Prune rings that are products of other rings
        # TODO - testing on molecules like phenalene shows that this is broken
        ring_sets = []
        for i in range(2, len(rings) + 1):
            for combo in combinations(rings, i):
                ring_set = set().union(*combo)
                ring_sets.append(ring_set)
        for ring in rings:
            ring_set = set(ring)
            if ring_set in ring_sets:
                _LOGGER.debug(f"Fused ring: {ring}")
            else:
                _LOGGER.debug(f"Unfused ring: {ring}")
                self.rings.add(ring)
        for ring in self.rings:
            for atom in ring:
                self.atoms[atom].num_rings += 1

    def read(self, mol2_file):
        """Routines for reading MOL2 file.

        :param mol2_file:  file-like object with MOL2 data
        """
        mol2_file = self.parse_atoms(mol2_file)
        mol2_file = self.parse_bonds(mol2_file)

    def parse_atoms(self, mol2_file):
        """Parse @<TRIPOS>ATOM section of file.

        :param mol2_file:  file-like object with MOL2 data
        :return:  file-like object advanced to bonds section
        :raises ValueError:  for bad MOL2 ATOM lines
        :raises TypeError:  for bad charge entries
        """
        # Skip material before atoms section
        for line in mol2_file:
            if "@<TRIPOS>ATOM" in line:
                break
            _LOGGER.debug(f"Skipping: {line.strip()}")
        duplicates = set()
        for line in mol2_file:
            line = line.strip()
            if not line:
                continue
            if "@<TRIPOS>BOND" in line:
                break
            words = line.split()
            if len(words) < 8:
                err = f"Bad entry in MOL2 file: {line}"
                raise ValueError(err)
            atom = Mol2Atom()
            atom.name = words[1]
            atom_type = words[5]
            type_parts = atom_type.split(".")
            type_parts[0] = type_parts[0].capitalize()
            if len(type_parts) == 2:
                type_parts[1] = type_parts[1].lower()
            elif len(type_parts) > 2:
                err = f"Invalid atom type: {atom_type}"
                raise ValueError(err)
            atom.type = ".".join(type_parts)
            atom.chain_id = "L"
            try:
                atom.serial = int(words[0])
                atom.res_name = words[7][:4]
                atom.res_seq = int(words[6])
                atom.x = float(words[2])
                atom.y = float(words[3])
                atom.z = float(words[4])
            except ValueError as exc:
                err = f"Error ({exc}) parsing atom line: {line}"
                raise ValueError(err)
            if len(line) > 8:
                try:
                    atom.mol2charge = float(words[8])
                except TypeError:
                    err = f"Unable to parse {words[8]} as charge in atom "
                    err += f"line: {line}"
                    _LOGGER.warning(err)
            if atom.name in self.atoms:
                duplicates.add(atom.name)
            else:
                self.atoms[atom.name] = atom
        if duplicates:
            raise KeyError(
                f"Found duplicate atoms names in MOL2 file: {duplicates}"
            )
        return mol2_file

    def parse_bonds(self, mol2_file):
        """Parse @<TRIPOS>BOND section of file.

        Atoms must already have been parsed.
        Also sets up torsions and rings.

        :param mol2_file:  file-like object with MOL2 data
        :return:  file-like object advanced to SUBSTRUCTURE section
        """
        atom_names = list(self.atoms.keys())
        for line in mol2_file:
            line = line.strip()
            if not line:
                continue
            if "@<TRIPOS>SUBSTRUCTURE" in line:
                break
            words = line.split()
            if len(words) < 4:
                err = f"Bond line too short: {line}"
                raise ValueError(err)
            bond_type = words[3]
            if bond_type == "1":
                bond_type = "single"
            elif bond_type == "2":
                bond_type = "double"
            elif bond_type == "3":
                bond_type = "triple"
            elif bond_type == "am":
                raise NotImplementedError(
                    "PDB2PQR does not currently support the amide (am) bond "
                    "type."
                )
            elif bond_type == "ar":
                bond_type = "aromatic"
            elif bond_type == "du":
                raise NotImplementedError(
                    "PDB2PQR does not currently support the dummy (du) bond "
                    "type."
                )
            elif bond_type == "un":
                raise NotImplementedError(
                    "PDB2PQR does not currently support the unknown (un) bond "
                    "type."
                )
            elif bond_type == "nc":
                raise NotImplementedError(
                    "PDB2PQR does not currently support the not-connected "
                    "(nc) bond type."
                )
            else:
                err = f"Unknown bond type: {bond_type}"
                raise ValueError(err)
            bond_id = int(words[0])
            atom_id1 = int(words[1])
            atom_id2 = int(words[2])
            atom_name1 = atom_names[atom_id1 - 1]
            atom1 = self.atoms[atom_name1]
            atom_name2 = atom_names[atom_id2 - 1]
            atom2 = self.atoms[atom_name2]
            bond = Mol2Bond(
                atom1=atom1, atom2=atom2, bond_type=bond_type, bond_id=bond_id
            )
            atom1.bonds.append(bond)
            atom1.bonded_atom_names.append(atom_name2)
            atom1.bonded_atoms.append(atom2)
            atom2.bonds.append(bond)
            atom2.bonded_atom_names.append(atom_name1)
            atom2.bonded_atoms.append(atom1)
            self.bonds.append(bond)
        self.set_torsions()
        self.set_rings()
        return mol2_file
